"""Given two list of blobs_in_video, one deduced from human groundtruth and the other
generated by the tracking algorithm, compares them and gives back some statistics

Crossing: crossings are a special case. We ...
"""
import logging
import os
import sys

import numpy as np

from idtrackerai import ListOfBlobs


def compare_tracking_against_groundtruth_no_gaps(
    number_of_animals,
    groundtruth,
    blobs_in_video_groundtruth,
    blobs_in_video,
    identities_dictionary_permutation,
):
    """
    blobs_in_video_groundtruth: cut groundtruth.blobs_in_video using start
    and end of the groundtruth object
    blobs_in_video: cut list_of_blobs.blobs_in_video using start and end of
    the groundtruth object
    """
    results = {
        "number_of_blobs_per_identity": {
            identity: 0 for identity in range(1, number_of_animals + 1)
        },
        "number_of_individuals_badly_assigned": {
            identity: 0 for identity in range(1, number_of_animals + 1)
        },
    }
    if identities_dictionary_permutation is not None:
        results["number_of_individuals_badly_interpolated"] = {
            identities_dictionary_permutation[
                identity
            ]: groundtruth.wrong_crossing_counter[identity]
            for identity in groundtruth.wrong_crossing_counter
        }
        results["number_of_individuals_unidentified"] = {
            identities_dictionary_permutation[
                identity
            ]: groundtruth.unidentified_individuals_counter[identity]
            for identity in groundtruth.wrong_crossing_counter
        }
    else:
        results["number_of_individuals_badly_interpolated"] = (
            groundtruth.wrong_crossing_counter
        )
        results["number_of_individuals_unidentified"] = (
            groundtruth.unidentified_individuals_counter
        )

    for blobs_in_frame_gt, blobs_in_frame in zip(
        blobs_in_video_groundtruth, blobs_in_video
    ):
        for blob_gt, blob in zip(blobs_in_frame_gt, blobs_in_frame):
            if identities_dictionary_permutation is not None:
                if isinstance(blob_gt.identity, int):
                    gt_identity = identities_dictionary_permutation[blob_gt.identity]
                elif isinstance(blob_gt.identity, list):
                    gt_identity = [
                        identities_dictionary_permutation[identity]
                        for identity in blob_gt.identity
                    ]
            else:
                gt_identity = blob_gt.identity

            if isinstance(gt_identity, int) and gt_identity != 0:
                results["number_of_blobs_per_identity"][gt_identity] += 1
            elif isinstance(gt_identity, list):
                for identity in gt_identity:
                    if identity != 0:
                        results["number_of_blobs_per_identity"][identity] += 1
            elif gt_identity is None:
                logging.debug("unidentified blobs")

            if (
                blob_gt.is_an_individual
                and not blob_gt.was_a_crossing
                and gt_identity != blob.assigned_identities
            ):
                results["number_of_individuals_badly_assigned"][gt_identity] += 1

    results["number_of_errors_in_all_blobs"] = {
        i: (
            results["number_of_individuals_badly_assigned"][i]
            + results["number_of_individuals_badly_interpolated"][i]
        )
        for i in range(1, number_of_animals + 1)
    }

    return results


def get_accuracy_wrt_groundtruth_no_gaps(
    video,
    groundtruth,
    blobs_in_video_groundtruth,
    blobs_in_video=None,
    first_frame_first_global_fragment=None,
):
    check_ground_truth_consistency(
        blobs_in_video_groundtruth, blobs_in_video, first_frame_first_global_fragment
    )
    identities_dictionary_permutation = get_permutation_of_identities(
        first_frame_first_global_fragment, blobs_in_video_groundtruth, blobs_in_video
    )
    number_of_animals = video.number_of_animals
    if blobs_in_video is None:
        blobs_in_video = blobs_in_video_groundtruth
    results = compare_tracking_against_groundtruth_no_gaps(
        number_of_animals,
        groundtruth,
        blobs_in_video_groundtruth,
        blobs_in_video,
        identities_dictionary_permutation,
    )

    accuracies = {
        "individual_accuracy": {
            i: (
                1.0
                - results["number_of_errors_in_all_blobs"][i]
                / results["number_of_blobs_per_identity"][i]
            )
            for i in range(1, number_of_animals + 1)
        }
    }
    accuracies["accuracy"] = np.mean(list(accuracies["individual_accuracy"].values()))
    logging.info(accuracies)
    logging.info(results)
    return accuracies, results


def compare_tracking_against_groundtruth(
    number_of_animals,
    blobs_in_video_groundtruth,
    blobs_in_video,
    identities_dictionary_permutation,
):
    # create dictionary to store eventual corrections made by the user
    results = {
        "number_of_blobs_per_identity": {i: 0 for i in range(1, number_of_animals + 1)},
        "sum_individual_P2": {i: 0 for i in range(1, number_of_animals + 1)},
        "number_of_assigned_blobs_per_identity": {
            i: 0 for i in range(1, number_of_animals + 1)
        },
        "number_of_blobs_assigned_during_accumulation_per_identity": {
            i: 0 for i in range(1, number_of_animals + 1)
        },
        "number_of_blobs_after_accumulation_per_identity": {
            i: 0 for i in range(1, number_of_animals + 1)
        },
        "number_of_errors_in_all_blobs": {
            i: 0 for i in range(1, number_of_animals + 1)
        },
        "number_of_errors_in_assigned_blobs": {
            i: 0 for i in range(1, number_of_animals + 1)
        },
        "number_of_errors_in_blobs_assigned_during_accumulation": {
            i: 0 for i in range(1, number_of_animals + 1)
        },
        "number_of_errors_in_blobs_after_accumulation": {
            i: 0 for i in range(1, number_of_animals + 1)
        },
        "number_of_errors_in_blobs_assigned_after_accumulation": {
            i: 0 for i in range(1, number_of_animals + 1)
        },
        "number_of_individual_blobs": 0,
        "number_of_crossing_blobs": 0,
        "number_of_crossings_blobs_assigned_as_individuals": 0,
        "frames_with_identity_errors": [],
        "fragment_identifiers_with_identity_errors": [],
        "frames_with_crossing_errors": [],
        "fragment_identifiers_with_crossing_errors": [],
        "frames_with_zeros_in_groundtruth": [],
        "number_of_crossing_fragments": 0,
        "fragments_identifiers_of_crossings": [],
    }

    for groundtruth_blobs_in_frame, blobs_in_frame in zip(
        blobs_in_video_groundtruth, blobs_in_video
    ):
        for groundtruth_blob, blob in zip(groundtruth_blobs_in_frame, blobs_in_frame):
            if identities_dictionary_permutation is not None:
                gt_identity = identities_dictionary_permutation[
                    groundtruth_blob.identity
                ]
            else:
                gt_identity = groundtruth_blob.identity

            if (
                groundtruth_blob.is_an_individual
                and gt_identity != -1
                and not groundtruth_blob.was_a_crossing
            ):  # we are not considering crossing or failures of the model area
                results["number_of_individual_blobs"] += 1
                if gt_identity == 0:
                    results["frames_with_zeros_in_groundtruth"].append(
                        groundtruth_blob.frame_number
                    )
                else:
                    try:
                        if (
                            blob.assigned_identities[0] != 0
                            and blob.identities_corrected_closing_gaps is None
                        ):  # we only consider P2 for non interpolated blobs
                            results["sum_individual_P2"][
                                gt_identity
                            ] += blob._P2_vector[gt_identity - 1]
                    except IndexError:
                        logging.debug("P2_vector %s" % str(blob._P2_vector))
                        logging.debug("individual %s" % str(blob.is_an_individual))
                        logging.debug(
                            "fragment identifier %s" % str(blob.fragment_identifier)
                        )
                    results["number_of_blobs_per_identity"][gt_identity] += 1
                    results["number_of_assigned_blobs_per_identity"][gt_identity] += (
                        1 if blob.assigned_identities[0] != 0 else 0
                    )
                    results[
                        "number_of_blobs_assigned_during_accumulation_per_identity"
                    ][gt_identity] += (1 if blob.used_for_training else 0)
                    results["number_of_blobs_after_accumulation_per_identity"][
                        gt_identity
                    ] += (1 if not blob.used_for_training else 0)
                    if gt_identity != blob.assigned_identities[0]:
                        results["number_of_errors_in_all_blobs"][gt_identity] += 1
                        results["number_of_errors_in_blobs_after_accumulation"][
                            gt_identity
                        ] += (1 if not blob.used_for_training else 0)
                        if blob.assigned_identities[0] != 0:
                            results["number_of_errors_in_assigned_blobs"][
                                gt_identity
                            ] += 1
                            results[
                                "number_of_errors_in_blobs_assigned_during_accumulation"
                            ][gt_identity] += (1 if blob.used_for_training else 0)
                            results[
                                "number_of_errors_in_blobs_assigned_after_accumulation"
                            ][gt_identity] += (1 if not blob.used_for_training else 0)
                        if (
                            blob.fragment_identifier
                            not in results["fragment_identifiers_with_identity_errors"]
                        ):
                            results["frames_with_identity_errors"].append(
                                blob.frame_number
                            )
                            results["fragment_identifiers_with_identity_errors"].append(
                                blob.fragment_identifier
                            )

            elif groundtruth_blob.is_a_crossing or gt_identity == -1:
                if (
                    blob.fragment_identifier
                    not in results["fragments_identifiers_of_crossings"]
                ):
                    results["fragments_identifiers_of_crossings"].append(
                        blob.fragment_identifier
                    )
                    results["number_of_crossing_fragments"] += 1
                results["number_of_crossing_blobs"] += 1
                results["number_of_crossings_blobs_assigned_as_individuals"] += (
                    1 if blob.is_an_individual else 0
                )
                if blob.is_an_individual and (
                    blob.fragment_identifier
                    not in results["fragment_identifiers_with_crossing_errors"]
                ):
                    results["frames_with_crossing_errors"].append(blob.frame_number)
                    results["fragment_identifiers_with_crossing_errors"].append(
                        blob.fragment_identifier
                    )

    return results


def check_ground_truth_consistency(
    blobs_in_video_groundtruth, blobs_in_video, first_frame_first_global_fragment
):
    if (
        first_frame_first_global_fragment is not None
        and first_frame_first_global_fragment > len(blobs_in_video_groundtruth)
    ):
        raise ValueError(
            "The first_frame_first_global_fragment is bigger than "
            "the length of the groundtruth video"
        )

    if len(blobs_in_video_groundtruth) != len(blobs_in_video):
        raise ValueError(
            "Cannot compute the accuracy from a list of blobs with different "
            "length than the blobs_in_video_groundtruth"
        )

    for blobs_in_frame_gt, blobs_in_frame in zip(
        blobs_in_video_groundtruth, blobs_in_video
    ):
        if len(blobs_in_frame) != len(blobs_in_frame_gt):
            raise ValueError(
                "Cannot compute the accuracy form a list of blobs with "
                "different blobs per frame than the blobs_in_video_groundtruth"
            )


def get_permutation_of_identities(
    first_frame_first_global_fragment, blobs_in_video_groundtruth, blobs_in_video
):
    if first_frame_first_global_fragment is not None:
        groundtruth_identities_in_first_frame = [
            blob.identity
            for blob in blobs_in_video_groundtruth[first_frame_first_global_fragment]
        ]
        identities_in_first_frame = [
            blob.identity for blob in blobs_in_video[first_frame_first_global_fragment]
        ]
        logging.debug(
            "groundtruth identities in first frame %s"
            % str(groundtruth_identities_in_first_frame)
        )
        logging.debug("identities in first frame %s" % str(identities_in_first_frame))

        identities_dictionary_permutation = {
            groundtruth_identity: identity
            for identity, groundtruth_identity in zip(
                identities_in_first_frame, groundtruth_identities_in_first_frame
            )
        }
    else:
        identities_dictionary_permutation = None

    return identities_dictionary_permutation


def get_accuracy_wrt_groundtruth(
    video,
    blobs_in_video_groundtruth,
    blobs_in_video=None,
    first_frame_first_global_fragment=None,
):
    check_ground_truth_consistency(
        blobs_in_video_groundtruth, blobs_in_video, first_frame_first_global_fragment
    )
    identities_dictionary_permutation = get_permutation_of_identities(
        first_frame_first_global_fragment, blobs_in_video_groundtruth, blobs_in_video
    )

    number_of_animals = video.number_of_animals
    if blobs_in_video is None:
        blobs_in_video = blobs_in_video_groundtruth
    results = compare_tracking_against_groundtruth(
        number_of_animals,
        blobs_in_video_groundtruth,
        blobs_in_video,
        identities_dictionary_permutation,
    )
    if len(results["frames_with_zeros_in_groundtruth"]) == 0:
        accuracies = {
            "percentage_of_unoccluded_images": results["number_of_individual_blobs"] / (
                results["number_of_individual_blobs"]
                + results["number_of_crossing_blobs"]
            ),
            "individual_P2_in_validated_part": {
                i: (
                    results["sum_individual_P2"][i]
                    / results["number_of_blobs_per_identity"][i]
                )
                for i in range(1, number_of_animals + 1)
            },
            "mean_individual_P2_in_validated_part": np.sum(
                list(results["sum_individual_P2"].values())
            ) / np.sum(list(results["number_of_blobs_per_identity"].values())),
            "individual_accuracy": {
                i: (
                    1
                    - results["number_of_errors_in_all_blobs"][i]
                    / results["number_of_blobs_per_identity"][i]
                )
                for i in range(1, number_of_animals + 1)
            },
            "accuracy": 1.0 - np.sum(
                list(results["number_of_errors_in_all_blobs"].values())
            ) / np.sum(list(results["number_of_blobs_per_identity"].values())),
            "individual_accuracy_assigned": {
                i: (
                    1
                    - results["number_of_errors_in_assigned_blobs"][i]
                    / results["number_of_assigned_blobs_per_identity"][i]
                )
                for i in range(1, number_of_animals + 1)
            },
            "accuracy_assigned": 1.0 - np.sum(
                list(results["number_of_errors_in_assigned_blobs"].values())
            ) / np.sum(list(results["number_of_assigned_blobs_per_identity"].values())),
            "individual_accuracy_in_accumulation": {
                i: (
                    1
                    - results["number_of_errors_in_blobs_assigned_during_accumulation"][
                        i
                    ]
                    / results[
                        "number_of_blobs_assigned_during_accumulation_per_identity"
                    ][i]
                )
                for i in range(1, number_of_animals + 1)
            },
            "accuracy_in_accumulation": 1.0 - np.sum(
                list(
                    results[
                        "number_of_errors_in_blobs_assigned_during_accumulation"
                    ].values()
                )
            ) / np.sum(
                list(
                    results[
                        "number_of_blobs_assigned_during_accumulation_per_identity"
                    ].values()
                )
            ),
            "individual_accuracy_after_accumulation": {},
        }
        for i in range(1, number_of_animals + 1):
            if results["number_of_blobs_after_accumulation_per_identity"][i] != 0:
                accuracies["individual_accuracy_after_accumulation"][i] = (
                    1
                    - results["number_of_errors_in_blobs_after_accumulation"][i]
                    / results["number_of_blobs_after_accumulation_per_identity"][i]
                )
            else:
                accuracies["individual_accuracy_after_accumulation"][i] = None
        if (
            np.sum(
                list(
                    results["number_of_blobs_after_accumulation_per_identity"].values()
                )
            )
            != 0
        ):
            accuracies["accuracy_after_accumulation"] = 1.0 - np.sum(
                list(results["number_of_errors_in_blobs_after_accumulation"].values())
            ) / np.sum(
                list(
                    results["number_of_blobs_after_accumulation_per_identity"].values()
                )
            )
        else:
            accuracies["accuracy_after_accumulation"] = None
        if results["number_of_crossing_blobs"] != 0:
            accuracies["crossing_detector_accuracy"] = (
                1.0
                - results["number_of_crossings_blobs_assigned_as_individuals"]
                / results["number_of_crossing_blobs"]
            )
        else:
            accuracies["crossing_detector_accuracy"] = None
        logging.info("accuracies %s" % str(accuracies))
        logging.info(
            "number of crossing fragments in ground truth interval: %i"
            % results["number_of_crossing_fragments"]
        )
        logging.info(
            "number of crossing blobs in ground truth interval: %i"
            % results["number_of_crossing_blobs"]
        )
        return accuracies, results

    logging.info(
        "there are fish with 0 identity in frame %s"
        % str(results["frames_with_zeros_in_groundtruth"])
    )
    return None, results


def compute_and_save_session_accuracy_wrt_groundtruth(video, groundtruth_type=None):
    logging.info("loading list_of_blobs")
    if groundtruth_type == "normal":
        list_of_blobs = ListOfBlobs.load(video, video.blobs_path)
    elif groundtruth_type == "interpolated":
        list_of_blobs = ListOfBlobs.load(video, video.blobs_path_interpolated)
    elif groundtruth_type == "no_gaps":
        list_of_blobs = ListOfBlobs.load(video, video.blobs_no_gaps_path)
    # select ground truth file
    logging.info("loading groundtruth")
    if groundtruth_type in ("normal", "interpolated"):
        groundtruth_path = video.ground_truth_path
    elif groundtruth_type == "no_gaps":
        groundtruth_path = os.path.join(
            video.video_folder, "_groundtruth_with_crossing_identified.npy"
        )
    groundtruth = np.load(groundtruth_path, allow_pickle=True).item()
    blobs_in_video_groundtruth = groundtruth.blobs_in_video[
        groundtruth.start : groundtruth.end
    ]
    blobs_in_video = list_of_blobs.blobs_in_video[groundtruth.start : groundtruth.end]
    logging.info("computing groundtruth")
    if groundtruth_type in ("normal", "interpolated"):
        accuracies, results = get_accuracy_wrt_groundtruth(
            video, blobs_in_video_groundtruth, blobs_in_video
        )
    elif groundtruth_type == "no_gaps":
        accuracies, results = get_accuracy_wrt_groundtruth_no_gaps(
            video, groundtruth, blobs_in_video_groundtruth, blobs_in_video
        )
    if accuracies is not None:
        logging.info("saving accuracies in video")
        video.gt_start_end = (groundtruth.start, groundtruth.end)
        if groundtruth_type == "normal":
            video.gt_accuracy = accuracies
            video.gt_results = results
        elif groundtruth_type == "interpolated":
            video.gt_accuracy_interpolated = accuracies
            video.gt_results_interpolated = results
        elif groundtruth_type == "no_gaps":
            video.gt_accuracy_no_gaps = accuracies
            video.gt_results_no_gaps = results
        video.save()


if __name__ == "__main__":
    from idtrackerai.utils.GUI_utils import selectDir

    groundtruth_type = sys.argv[1]
    # select blobs_in_video list tracked to compare against ground truth
    session_path = selectDir("./")  # select path to video
    video_object_path = os.path.join(session_path, "video_object.npy")
    logging.info("loading video object")
    video = np.load(video_object_path, allow_pickle=True).item(0)
    compute_and_save_session_accuracy_wrt_groundtruth(video, groundtruth_type)



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import logging

import cv2

from idtrackerai import ListOfBlobs, Video
from idtrackerai.utils import CustomError, create_dir

from .segmentation import compute_background, segment


def animals_detection_API(video: Video):
    """
    This class generates a ListOfBlobs object and updates the video
    object with information about the process.

    Parameters
    ----------
    video: Video
        An instance of the class :class:`~idtrackerai.video.Video`.

    Attributes
    ----------
    video: Video
    list_of_blobs: ListOfBlobs
    detection_parameters: Dict

    See Also
    --------
    :class:`~idtrackerai.list_of_blobs.ListOfBlobs`
    """
    video.detect_animals_timer.start()
    create_dir(video.segmentation_data_folder, remove_existing=True)

    bkg_model = video.bkg_model
    if video.use_bkg:
        if bkg_model is None:
            bkg_model = compute_background(
                video.video_paths,
                video.episodes,
                video.number_of_frames_for_background,
                video.background_subtraction_stat,
            )
            video.bkg_model = bkg_model
        else:
            logging.info("Using previously computed background model from GUI")
    else:
        bkg_model = None
        logging.info("No background model computed")

    detection_parameters = {
        "intensity_ths": video.intensity_ths,
        "area_ths": video.area_ths,
        "ROI_mask": video.ROI_mask,
        "bkg_model": bkg_model,
        "resolution_reduction": video.resolution_reduction,
    }

    if video.resolution_reduction != 1 and bkg_model is not None:
        detection_parameters["bkg_model"] = cv2.resize(
            bkg_model,
            None,  # type: ignore
            fx=video.resolution_reduction,
            fy=video.resolution_reduction,
            interpolation=cv2.INTER_AREA,
        )

    # Main call
    blobs_in_video = segment(
        detection_parameters,
        video.episodes,
        video.segmentation_data_folder / "blobs_bbox_images.hdf5",
        video.video_paths,
        video.number_of_frames,
        video.number_of_parallel_workers,
    )

    list_of_blobs = ListOfBlobs(blobs_in_video)
    assert len(list_of_blobs) == video.number_of_frames
    logging.info(f"{list_of_blobs.number_of_blobs} detected blobs in total")

    if video.n_animals > 0:
        check_segmentation(video, list_of_blobs)

    video.detect_animals_timer.finish()
    return list_of_blobs


def check_segmentation(video: Video, list_of_blobs: ListOfBlobs):
    """
    idtracker.ai is designed to work under the assumption that all the
    detected blobs are animals. In the frames where the number of
    detected blobs is higher than the number of animals in the video, it is
    likely that some blobs do not represent animals. In this scenario
    idtracker.ai might misbehave. This method allows to check such
    condition.
    """
    logging.info("Checking segmentation")

    n_frames_with_all_visible = sum(
        len(blobs) == video.n_animals for blobs in list_of_blobs.blobs_in_video
    )

    if n_frames_with_all_visible == 0:
        raise CustomError(
            "There is not any frames where the number of blobs is equal "
            "to the number of animals stated by the user. Idtracker.ai "
            "needs those frame to work"
        )

    error_frames = [
        frame
        for frame, blobs in enumerate(list_of_blobs.blobs_in_video)
        if len(blobs) > video.n_animals
    ]

    n_error_frames = len(error_frames)
    logging.log(
        logging.WARNING if n_error_frames else logging.INFO,
        f"There are {n_error_frames} frames with more blobs than animals",
    )
    video.number_of_error_frames = n_error_frames

    if n_error_frames:
        logging.warning(
            "This can be detrimental for the proper functioning of the system"
        )
        if n_error_frames < 25:
            logging.warning(f"Frames with more blobs than animals: {error_frames}")
        else:
            logging.warning(
                "Too many frames with more blobs than animals "
                "for printing their indices in log"
            )

        output_path = video.session_folder / "inconsistent_frames.csv"
        logging.info(
            f"Saving indices of frames with more blobs than animals in {output_path}"
        )
        output_path.write_text("\n".join(map(str, error_frames)))

        if video.check_segmentation:
            list_of_blobs.save(video.blobs_path)
            raise CustomError(
                f"Check_segmentation is {True}, exiting...\n"
                "Please readjust the segmentation parameters and track again"
            )
        logging.info(f"Check_segmentation is {False}, ignoring the above errors")



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)

import logging
import os
import sys

import numpy as np

from idtrackerai import ListOfBlobs, ListOfFragments

ATTRIBUTES_TO_COPY = (
    "identities",
    "centroids",
    "assigned_identities",
    "assigned_centroids",
    "used_for_training",
    "accumulation_step",
    "centroid",
    "pixels",
    "frame_number",
    "is_an_individual",
    "is_a_crossing",
    "was_a_crossing",
    "blob_index",
    "fragment_identifier",
)


class GroundTruthBlob:
    """Lighter blob objects.
    Attributes:
        identity (preferring the one assigned by the user, if it is not None)
        centroid
        pixels (pixels is stored to check the groundtruth in crossings)
    """

    def __init__(self, blob):
        self.attributes = ATTRIBUTES_TO_COPY
        self.set_attributes(blob)

    def set_attributes(self, blob):
        for attribute in self.attributes:
            if attribute == "identities":
                setattr(self, attribute, blob.final_identities)
            elif attribute == "centroids":
                setattr(self, attribute, blob.final_centroids)
            else:
                setattr(self, attribute, getattr(blob, attribute))

    @property
    def gt_identity(self):
        if hasattr(self, "identity"):
            return self.identity

        if len(self.identities) == 1:
            return self.identities[0]
        return -1


class GroundTruth:
    def __init__(self, video, blobs_in_video, start, end):
        self.video = video
        self.blobs_in_video = blobs_in_video
        self.start = start
        self.end = end

    def save(self, name=""):
        if name == "":
            path_to_save_groundtruth = video.ground_truth_path
        else:
            path_to_save_groundtruth = os.path.join(
                self.video.video_folder, "_groundtruth_" + name + ".npy"
            )
        logging.info("saving ground truth at %s" % path_to_save_groundtruth)
        np.save(path_to_save_groundtruth, self)
        logging.info("done")


def generate_groundtruth(
    video,
    blobs_in_video=None,
    start=None,
    end=None,
    wrong_crossing_counter=None,
    unidentified_individuals_counter=None,
    save_gt=True,
):
    """Generates a list of light blobs_in_video, given a video object corresponding to a
    tracked video
    """
    logging.info("Generating ground truth file")
    # make sure the video has been succesfully tracked
    blobs_in_video_groundtruth = []

    for blobs_in_frame in blobs_in_video:
        blobs_in_frame_groundtruth = []

        for blob in blobs_in_frame:
            gt_blob = GroundTruthBlob(blob)
            blobs_in_frame_groundtruth.append(gt_blob)

        blobs_in_video_groundtruth.append(blobs_in_frame_groundtruth)

    groundtruth = GroundTruth(
        video=video, blobs_in_video=blobs_in_video_groundtruth, start=start, end=end
    )
    groundtruth.wrong_crossing_counter = wrong_crossing_counter
    groundtruth.unidentified_individuals_counter = unidentified_individuals_counter
    if save_gt:
        groundtruth.save()
    return groundtruth


if __name__ == "__main__":
    session_path = sys.argv[1]  # select path to video
    video_path = os.path.join(session_path, "video_object.npy")
    video = np.load(video_path, allow_pickle=True).item()
    start = input(
        "GroundTruth (start)"
        "Input the starting frame for the interval "
        "in which the video has been validated"
    )
    end = input(
        "GroundTruth (end)"
        "Input the ending frame for the interval "
        "in which the video has been validated"
    )
    # read blob list from video
    fragments_path = os.path.join(session_path, "preprocessing", "fragments.npy")
    blobs_path = os.path.join(session_path, "preprocessing", "blobs_collection.npy")
    list_of_fragments = ListOfFragments.load(fragments_path)
    list_of_blobs = ListOfBlobs.load(video, blobs_path)
    list_of_blobs.update_from_list_of_fragments(list_of_fragments.fragments)
    groundtruth = generate_groundtruth(
        video, list_of_blobs.blobs_in_video, int(start), int(end)
    )



import warnings

import numpy as np
from qtpy.QtCore import Qt, Signal  # type: ignore
from qtpy.QtGui import QKeyEvent
from qtpy.QtWidgets import (
    QAbstractItemView,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QTableWidget,
    QTableWidgetItem,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from idtrackerai import ListOfBlobs
from idtrackerai_GUI_tools import LabeledSlider, key_event_modifier


class CustomTableWidget(QTableWidget):
    def keyPressEvent(self, e: QKeyEvent):
        event = key_event_modifier(e)
        if event is not None:
            super().keyPressEvent(event)

    def keyReleaseEvent(self, e: QKeyEvent):
        event = key_event_modifier(e)
        if event is not None:
            super().keyReleaseEvent(event)


class CustomTableWidgetItem(QTableWidgetItem):
    def __init__(self, value: str | int):
        super().__init__("" if value == -1 else str(value))
        self.setData(Qt.ItemDataRole.UserRole, value)
        self.setTextAlignment(Qt.AlignmentFlag.AlignCenter)

    def __lt__(self, other: QTableWidgetItem) -> bool:
        return self.data(Qt.ItemDataRole.UserRole) < other.data(
            Qt.ItemDataRole.UserRole
        )


class ErrorsExplorer(QWidget):
    go_to_error = Signal(str, int, int, np.ndarray, int)
    # kind, start, end, where, id

    def __init__(self):
        super().__init__()
        self.table = CustomTableWidget(1, 4)
        horizontalHeader = self.table.horizontalHeader()
        horizontalHeader.setSectionResizeMode(QHeaderView.ResizeMode.Stretch)
        horizontalHeader.setSectionResizeMode(
            0, QHeaderView.ResizeMode.ResizeToContents
        )
        horizontalHeader.setMinimumSectionSize(10)
        verticalHeader = self.table.verticalHeader()
        verticalHeader.setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)
        verticalHeader.setVisible(False)
        self.table.setAlternatingRowColors(True)
        self.table.setWordWrap(False)
        self.table.setSelectionBehavior(QTableWidget.SelectionBehavior.SelectRows)
        self.table.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self.table.setHorizontalHeaderLabels(["Type", "Id", "Start", "Length"])
        self.table.setSortingEnabled(True)
        self.table.currentCellChanged.connect(self.cell_clicked)
        self.table.cellDoubleClicked.connect(self.cell_clicked)

        long_jumps_row = QHBoxLayout()
        self.jumps_th_label = QLabel("Jumps threshold")
        long_jumps_row.addWidget(self.jumps_th_label)
        self.jumps_th = LabeledSlider(self, 5, 10)
        self.jumps_th.setValue(8)
        self.jumps_th.valueChanged.connect(self.update_list_of_errors)
        long_jumps_row.addWidget(self.jumps_th)
        self.reset_jumps = QToolButton()
        self.reset_jumps.setText("Reset")
        self.reset_jumps.clicked.connect(lambda: self.non_accepted_jumps.fill(True))
        self.reset_jumps.clicked.connect(self.update_list_of_errors)
        long_jumps_row.addWidget(self.reset_jumps)

        layout = QVBoxLayout()
        left_widget = QWidget()
        left_widget.setLayout(layout)
        errors_header = QHBoxLayout()
        self.update_btn = QToolButton()
        self.update_btn.setIcon(
            self.style().standardIcon(self.style().StandardPixmap.SP_BrowserReload)
        )
        self.update_btn.setShortcut(Qt.Key.Key_U)
        self.update_btn.clicked.connect(self.update_list_of_errors)
        self.left_label = QLabel()
        errors_header.addWidget(self.left_label)
        errors_header.addWidget(self.update_btn)
        layout.addLayout(errors_header)
        layout.addWidget(self.table)
        layout.addLayout(long_jumps_row)
        self.setLayout(layout)
        layout.setContentsMargins(0, 0, 0, 0)

        self.trajectories: np.ndarray
        self.unidentified: np.ndarray
        self.duplicated: np.ndarray
        self.non_accepted_jumps: np.ndarray
        self.in_tracking_interval: np.ndarray

    def cell_clicked(self, row: int, col: int):
        if row < 0 or col < 0:
            return

        kind = self.table.item(row, 0).data(Qt.ItemDataRole.UserRole)
        identity = self.table.item(row, 1).data(Qt.ItemDataRole.UserRole)
        start: int = self.table.item(row, 2).data(Qt.ItemDataRole.UserRole)
        length = self.table.item(row, 3).data(Qt.ItemDataRole.UserRole)
        self.selected_error = kind, identity, start, length

        where = self.trajectories[start]
        if kind in ("Jump", "Miss id"):
            if start > 0:
                where = self.trajectories[start - 1 : start + length + 1, identity - 1]
            else:
                where = self.trajectories[start : start + length + 1, identity - 1]
        elif kind == "No id":
            for blob in self.list_of_blobs.blobs_in_video[start]:
                for blob_id, centroid in blob.final_ids_and_centroids:
                    if blob_id in (None, 0):
                        where = np.asarray(centroid)
                        break
        elif kind == "Dupl":
            duplicated_centroids = []
            for blob in self.list_of_blobs.blobs_in_video[start]:
                for blob_id, centroid in blob.final_ids_and_centroids:
                    if blob_id == identity:
                        duplicated_centroids.append(centroid)
            where = np.asarray(duplicated_centroids)
        else:
            raise ValueError(kind)

        self.go_to_error.emit(kind, start, length, where, identity)

    def set_references(
        self,
        traj: np.ndarray,
        unidentified: np.ndarray,
        duplicated: np.ndarray,
        blobs: ListOfBlobs,
        tracking_intervals: list[list[int]] | None,
    ):
        self.trajectories = traj
        self.unidentified = unidentified
        self.duplicated = duplicated
        self.list_of_blobs = blobs
        self.non_accepted_jumps = np.ones((traj.shape[0] - 1, traj.shape[1]), bool)
        if tracking_intervals is None:
            self.in_tracking_interval = np.ones(traj.shape[0], bool)
        else:
            self.in_tracking_interval = np.zeros(traj.shape[0], bool)
            for start, end in tracking_intervals:
                self.in_tracking_interval[start:end] = True

        self.update_list_of_errors()

    def accepted_interpolation(self):
        if not hasattr(self, "selected_error"):
            return
        kind, identity, start, length = self.selected_error
        start -= 1
        if kind == "Jump":
            self.non_accepted_jumps[start : start + length, identity - 1] = False

    def getErrors(self) -> dict[str, list[tuple[int, np.ndarray, np.ndarray]]]:
        # TODO Add more errors (super-crossings)
        if np.isnan(self.trajectories).all():
            # not finished session
            return {}
        return {
            "Miss id": get_list_of_Trues_for_id(
                np.isnan(self.trajectories[..., 0]) & self.in_tracking_interval[:, None]
            ),
            "No id": [(-1,) + get_list_of_Trues(self.unidentified)],
            "Dupl": get_list_of_Trues_for_id(self.duplicated),
            "Jump": self.get_impossible_jumps(),
        }

    def update_list_of_errors(self):
        self.table.setSortingEnabled(False)
        self.table.setRowCount(0)
        for error_kind, errors_for_id in self.getErrors().items():
            for identity, starts, lengths in errors_for_id:
                for start, length in zip(starts, lengths):
                    self.table.insertRow(0)
                    self.table.setItem(0, 0, CustomTableWidgetItem(error_kind))
                    self.table.setItem(0, 1, CustomTableWidgetItem(identity))
                    self.table.setItem(0, 2, CustomTableWidgetItem(start))
                    self.table.setItem(0, 3, CustomTableWidgetItem(length))
        self.left_label.setText(f"List of errors ({self.table.rowCount()} errors)")
        self.table.setSortingEnabled(True)

    def get_impossible_jumps(self):
        speed = np.sqrt(np.sum(np.diff(self.trajectories, axis=0) ** 2, axis=-1))
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            try:
                mean, std = np.nanmean(speed), np.nanstd(speed)
            except RuntimeWarning:
                return []

        too_fast = (
            speed > (mean + self.jumps_th.value() * std)
        ) & self.non_accepted_jumps
        out = get_list_of_Trues_for_id(too_fast)
        for _id, start, _length in out:
            start += 1
        return out


def get_list_of_Trues(arr: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    """Returns the start and the length of every True cluster in an array

    Input: [0,1,0,1,1,0]
    Output [1,3], [1,2] (one cluster at 1 of length 1 and another at 3 of length 2)
    """
    where = arr.nonzero()[0]
    is_edge = np.diff(where, prepend=-np.inf, append=np.inf) > 1
    starts = where[is_edge[:-1]]
    ends = where[is_edge[1:]]
    return starts, ends - starts + 1


def get_list_of_Trues_for_id(
    data: np.ndarray,
) -> list[tuple[int, np.ndarray, np.ndarray]]:
    return [
        (fish_id + 1,) + get_list_of_Trues(data[:, fish_id])
        for fish_id in range(data.shape[1])
    ]



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import logging

from idtrackerai import Blob, ListOfBlobs, ListOfFragments, ListOfGlobalFragments, Video
from idtrackerai.utils import track


def fragmentation_API(
    video: Video, list_of_blobs: ListOfBlobs
) -> tuple[ListOfFragments, ListOfGlobalFragments]:
    video.fragmentation_timer.start()

    compute_fragment_identifier_and_blob_index(
        list_of_blobs.blobs_in_video,
        max(video.n_animals, list_of_blobs.maximum_number_of_blobs),
    )

    list_of_fragments = ListOfFragments.from_fragmented_blobs(
        list_of_blobs.all_blobs, video.n_animals, video.id_images_file_paths
    )
    logging.info(
        f"{list_of_fragments.number_of_fragments} Fragments in total, "
        f"{list_of_fragments.number_of_individual_fragments} individuals and "
        f"{list_of_fragments.number_of_crossing_fragments} crossings"
    )

    list_of_global_fragments = ListOfGlobalFragments.from_fragments(
        list_of_blobs.blobs_in_video, list_of_fragments.fragments, video.n_animals
    )
    list_of_fragments.manage_accumulable_non_accumulable_fragments(
        list_of_global_fragments.global_fragments,
        list_of_global_fragments.non_accumulable_global_fragments,
    )

    video.fragmentation_timer.finish()
    return list_of_fragments, list_of_global_fragments


def compute_fragment_identifier_and_blob_index(
    blobs_in_video: list[list[Blob]], number_of_animals: int
) -> None:
    """Associates a unique fragment identifier to individual blobs
    connected with its next and previous blobs.

    Blobs must be connected and classified as individuals or crossings.

    Parameters
    ----------
    number_of_animals : int
        Number of animals to be tracked as defined by the user
    """
    frame_id = 0
    possible_blob_indices = set(range(number_of_animals))

    for blobs_in_frame in track(blobs_in_video, "Fragmenting blobs"):
        missing_blob_indices = possible_blob_indices.difference(
            blob.blob_index for blob in blobs_in_frame
        )

        for blob in blobs_in_frame:
            if blob.fragment_identifier != -1:
                continue

            blob.fragment_identifier = frame_id
            if blob.is_an_individual:
                blob_index = missing_blob_indices.pop()
                blob.blob_index = blob_index
                while (
                    blob.n_next == 1
                    and blob.next[0].n_previous == 1
                    and blob.next[0].is_an_individual
                ):
                    blob = blob.next[0]
                    blob.fragment_identifier = frame_id
                    blob.blob_index = blob_index

            elif blob.is_a_crossing:
                while (
                    blob.n_next == 1
                    and blob.next[0].n_previous == 1
                    and blob.next[0].is_a_crossing
                ):
                    blob = blob.next[0]
                    blob.fragment_identifier = frame_id

            frame_id += 1



import logging
import platform
from pathlib import Path

import numpy as np
from torch.utils.data import DataLoader
from torchvision import transforms
from torchvision.datasets.folder import VisionDataset

from idtrackerai import Blob
from idtrackerai.network import normalize
from idtrackerai.tracker.identity_dataset import duplicate_PCA_images
from idtrackerai.utils import conf, load_id_images, track


class CrossingDataset(VisionDataset):
    images: np.ndarray
    labels: np.ndarray

    def __init__(
        self,
        blobs_list: list[Blob] | dict[str, list[Blob]],
        id_images_file_paths: list[Path],
        scope,
        transform=None,
    ):
        super().__init__("", transform=transform)
        self.id_images_file_paths = id_images_file_paths
        self.blobs = blobs_list
        self.scope = scope
        self.get_data()

    def get_data(self):
        if isinstance(self.blobs, dict):
            logging.info(f"Generating crossing {self.scope} set.")
            crossings_images = self.get_images_indices(image_type="crossings")
            crossing_labels = np.ones(len(crossings_images), int)

            logging.info(f"Generating single individual {self.scope} set")
            individual_images = self.get_images_indices(image_type="individuals")
            individual_labels = np.zeros(len(individual_images), int)

            logging.info("Preparing images and labels")
            images_indices = crossings_images + individual_images
            self.images = load_id_images(self.id_images_file_paths, images_indices)
            self.images = np.expand_dims(self.images, axis=-1)

            self.labels = np.concatenate([crossing_labels, individual_labels], axis=0)

            if self.scope == "training":
                self.images, self.labels = duplicate_PCA_images(
                    self.images, self.labels
                )

        elif isinstance(self.blobs, list):
            images_indices = self.get_images_indices()
            self.images = load_id_images(self.id_images_file_paths, images_indices)
            self.images = np.expand_dims(self.images, axis=-1)
            self.labels = np.zeros((self.images.shape[0]))

    def get_images_indices(self, image_type: str = "") -> list[tuple[int, int]]:
        if image_type:
            assert isinstance(self.blobs, dict)
            blobs = self.blobs[image_type]
        else:
            assert isinstance(self.blobs, list)
            blobs = self.blobs

        return [(blob.id_image_index, blob.episode) for blob in blobs]

    def __len__(self):
        return len(self.images)

    def __getitem__(self, index):
        image = self.images[index]
        target = self.labels[index]
        if self.transform is not None:
            image = self.transform(image)
        return image, target


def get_train_validation_and_eval_blobs(
    blobs_in_video: list[list[Blob]],
    number_of_animals: int,
    ratio_validation: float = 0.1,
) -> tuple[dict[str, list[Blob]], dict[str, list[Blob]], list[Blob]]:
    """Given a list of blobs return 2 dictionaries (training_blobs, validation_blobs),
    and a list (toassign_blobs).

    :param list_of_blobs:
    :param ratio_validation:
    :return: training_blobs, validation_blobs, toassign_blobs
    """
    logging.info("Get list of blobs for training, validation and eval")

    individuals = []
    crossings = []
    toassign_blobs = []
    for blobs_in_frame in track(blobs_in_video, "First individual/crossing assignment"):
        in_a_global_fragment_core = len(blobs_in_frame) == number_of_animals
        for blob in blobs_in_frame:
            if in_a_global_fragment_core or blob.is_a_sure_individual():
                blob.used_for_training_crossings = True
                blob.is_an_individual = True
                individuals.append(blob)
            elif blob.is_a_sure_crossing():
                blob.used_for_training_crossings = True
                blob.is_an_individual = False
                crossings.append(blob)
            else:
                blob.used_for_training_crossings = False
                toassign_blobs.append(blob)

    logging.debug(
        f"{len(individuals)} individual, "
        f"{len(crossings)} crossing and "
        f"{len(toassign_blobs)} unknown blobs in total"
    )

    # Shuffle and make crossings and individuals even
    rng = np.random.default_rng()
    rng.shuffle(individuals)
    rng.shuffle(crossings)

    crossings = crossings[: conf.MAX_IMAGES_PER_CLASS_CROSSING_DETECTOR]
    individuals = individuals[: conf.MAX_IMAGES_PER_CLASS_CROSSING_DETECTOR]

    n_blobs_crossings = len(crossings)
    n_blobs_individuals = len(individuals)
    n_individual_blobs_validation = int(n_blobs_individuals * ratio_validation)
    n_crossing_blobs_validation = int(n_blobs_crossings * ratio_validation)

    # split training and validation
    validation_blobs = {
        "individuals": individuals[:n_individual_blobs_validation],
        "crossings": crossings[:n_crossing_blobs_validation],
    }

    training_blobs = {
        "individuals": individuals[n_individual_blobs_validation:],
        "crossings": crossings[n_crossing_blobs_validation:],
    }

    ratio_crossings = n_blobs_crossings / (n_blobs_crossings + n_blobs_individuals)
    training_blobs["weights"] = [ratio_crossings, 1 - ratio_crossings]

    logging.info(
        f"{len(training_blobs['individuals'])} individual and "
        f"{len(training_blobs['crossings'])} crossing blobs for training\n"
        f"{len(validation_blobs['individuals'])} individual and "
        f"{len(validation_blobs['crossings'])} crossing blobs for validation\n"
        f"{len(toassign_blobs)} blobs to test"
    )

    return training_blobs, validation_blobs, toassign_blobs


if platform.system() in ("Windows", "Darwin"):
    # Using multiprocessing in Windows and MacOS causes a
    # recursion limit error difficult to debug
    num_workers_train = 0
    num_workers_val = 0
else:
    num_workers_train = 1
    num_workers_val = 1


def get_training_data_loaders(
    id_images_file_paths: list[Path],
    train_blobs: dict[str, list[Blob]],
    val_blobs: dict[str, list[Blob]],
):
    logging.info("Creating training and validation data loaders")
    training_set = CrossingDataset(
        train_blobs,
        id_images_file_paths,
        scope="training",
        transform=transforms.Compose([transforms.ToTensor(), normalize]),
    )
    train_loader = DataLoader(
        training_set,
        batch_size=conf.BATCH_SIZE_DCD,
        shuffle=True,
        num_workers=num_workers_train,
        persistent_workers=num_workers_train > 0,
    )
    train_loader.num_classes = 2
    train_loader.image_shape = training_set[0][0].shape

    logging.info("Creating validation CrossingDataset")
    validation_set = CrossingDataset(
        val_blobs,
        id_images_file_paths,
        scope="validation",
        transform=transforms.Compose([transforms.ToTensor(), normalize]),
    )
    val_loader = DataLoader(
        validation_set,
        batch_size=conf.BATCH_SIZE_PREDICTIONS_DCD,
        num_workers=num_workers_val,
        persistent_workers=num_workers_val > 0,
    )
    val_loader.num_classes = 2
    val_loader.image_shape = validation_set[0][0].shape
    return train_loader, val_loader


def get_test_data_loader(id_images_file_paths: list[Path], test_blobs: list[Blob]):
    logging.info("Creating test CrossingDataset")
    test_set = CrossingDataset(
        test_blobs,
        id_images_file_paths,
        scope="test",
        transform=transforms.Compose([transforms.ToTensor(), normalize]),
    )
    test_loader = DataLoader(
        test_set,
        batch_size=conf.BATCH_SIZE_PREDICTIONS_DCD,
        num_workers=num_workers_val,
        persistent_workers=num_workers_val > 0,
    )
    test_loader.num_classes = 2
    test_loader.image_shape = test_set[0][0].shape
    return test_loader



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
def compare_tracked_individual_against_groundtruth(
    blobs_in_individual_groundtruth, individual_blobs, individual_groundtruth_id
):
    comparison_keys = ["accuracy", "frames_with_errors", "mistaken_identities"]
    comparison_info = {key: [] for key in comparison_keys}

    for blob_gt, blob in zip(blobs_in_individual_groundtruth, individual_blobs):
        if blob_gt.identity != blob.assigned_identity:
            comparison_info["frames_with_errors"].append(blob.frame_number)
            comparison_info["mistaken_identities"].append(blob.assigned_identity)

    comparison_info["accuracy"] = 1 - len(comparison_info["mistaken_identities"]) / len(
        blobs_in_individual_groundtruth
    )
    number_of_assigned_blobs = len(
        [
            blob
            for blob in blobs_in_individual_groundtruth
            if blob.assigned_identity != 0
        ]
    )
    number_of_mistaken_identified_blobs = len(
        [
            identity
            for identity in comparison_info["mistaken_identities"]
            if identity != 0
        ]
    )
    comparison_info["accuracy_assigned"] = (
        1 - number_of_mistaken_identified_blobs / number_of_assigned_blobs
    )
    comparison_info["id"] = individual_groundtruth_id
    return comparison_info


def check_groundtruth_consistency(
    blobs_in_individual_groundtruth,
    individual_groundtruth_id,
    individual_blobs,
    individual_id,
):
    non_matching_error = (
        "The length of the collections of the ground truth "
        "individual and the selected one do not match"
    )
    if individual_groundtruth_id != individual_id or len(
        blobs_in_individual_groundtruth
    ) != len(individual_blobs):
        raise ValueError(non_matching_error)


def get_individual_accuracy_wrt_groundtruth(
    blobs_in_individual_groundtruth, individual_blobs=None
):
    individual_groundtruth_id = blobs_in_individual_groundtruth[0].identity
    individual_id = individual_groundtruth_id
    if individual_blobs is None:
        individual_blobs = blobs_in_individual_groundtruth
    else:
        check_groundtruth_consistency(
            blobs_in_individual_groundtruth,
            individual_groundtruth_id,
            individual_blobs,
            individual_id,
        )
    return compare_tracked_individual_against_groundtruth(
        blobs_in_individual_groundtruth, individual_blobs, individual_groundtruth_id
    )



from typing import Optional

import numpy as np
from qtpy.QtCore import QEvent, QPoint, QPointF, Qt
from qtpy.QtGui import QKeyEvent, QPainterPath, QPalette, QResizeEvent
from qtpy.QtWidgets import QFrame, QLabel, QSizePolicy, QWidget
from superqt import QLabeledRangeSlider, QLabeledSlider
from superqt.sliders._labeled import LabelPosition

from idtrackerai.utils import get_vertices_from_label


class QHLine(QFrame):
    def __init__(self):
        super().__init__()
        self.setFrameShape(QFrame.Shape.HLine)
        self.setContentsMargins(10, 0, 10, 0)
        self.setEnabled(False)

    def changeEvent(self, event: QEvent):
        if event.type() == QEvent.Type.EnabledChange:
            self.setEnabled(False)


def build_ROI_patches_from_list(
    list_of_ROIs: list[str] | str | None,
    resolution_reduction: float,
    width: int,
    height: int,
) -> QPainterPath:
    path = QPainterPath()

    if list_of_ROIs is None:
        return path

    if isinstance(list_of_ROIs, str):
        list_of_ROIs = [list_of_ROIs]

    path.addRect(
        -0.5, -0.5, width * resolution_reduction, height * resolution_reduction
    )

    for line in list_of_ROIs:
        points = (get_vertices_from_label(line) * resolution_reduction + 0.5).astype(
            np.int32
        )
        path_i = QPainterPath(QPointF(*points[0]))
        for point in points[1:]:
            path_i.lineTo(*point)

        if line[0] == "+":
            path -= path_i.simplified()
        elif line[0] == "-":
            path += path_i.simplified()
        else:
            raise TypeError
    return path


class LabeledSlider(QLabeledSlider):
    def __init__(self, parent: QWidget, min, max):
        self.parent_widget = parent
        super().__init__(Qt.Orientation.Horizontal, parent)
        self.setRange(min, max)
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)
        self._label.setFocusPolicy(Qt.FocusPolicy.WheelFocus)
        self._label.valueChanged.connect(lambda val: self._slider.setValue(int(val)))

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() in (
            QEvent.Type.PaletteChange,
            QEvent.Type.EnabledChange,
            QEvent.Type.FontChange,
        ):
            style = (
                "QDoubleSpinBox{"
                + f"color: #{self.palette().color(QPalette.ColorRole.Text).rgba():x}"
                ";background:transparent; border: 0;"
                f" font-size:{self.font().pointSize()}pt"
                "}QDoubleSpinBox:!enabled{color: #"
                + f"{self.palette().color(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text).rgba():x}"
                ";}"
            )
            self._label.setStyleSheet(style)
            self._label._update_size()


class LabelRangeSlider(QLabeledRangeSlider):
    def __init__(
        self,
        min: int,
        max: int,
        parent: Optional[QWidget] = None,
        start_end_val: Optional[tuple[int, int]] = None,
        block_upper=True,
    ):
        self.parent_widget = parent
        super().__init__(Qt.Orientation.Horizontal, parent)
        self.block_upper = block_upper
        self.setRange(min, max)
        self.setValue(start_end_val or (min, max))
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)

        self._min_label.setReadOnly(True)
        if block_upper:
            self._max_label.setReadOnly(True)
        else:
            self._max_label.setFocusPolicy(Qt.FocusPolicy.WheelFocus)

        self._handle_labels[0].valueChanged.connect(
            lambda val: self._slider.setSliderPosition(int(val), 0)
        )
        self._handle_labels[1].valueChanged.connect(
            lambda val: self._slider.setSliderPosition(int(val), 1)
        )

        for handle in self._handle_labels:
            handle.setFocusPolicy(Qt.FocusPolicy.WheelFocus)

    def _reposition_labels(self):
        """Overriding superqt method to remove the last label.clearFocus() call"""
        if (
            not self._handle_labels
            or self._handle_label_position == LabelPosition.NoLabel
        ):
            return

        horizontal = self.orientation() == Qt.Orientation.Horizontal
        labels_above = self._handle_label_position == LabelPosition.LabelsAbove

        last_edge = None
        for i, label in enumerate(self._handle_labels):
            rect = self._slider._handleRect(i)
            dx = -label.width() / 2
            dy = -label.height() / 2
            if labels_above:
                if horizontal:
                    dy *= 3
                else:
                    dx *= -1
            else:
                if horizontal:
                    dy *= -1
                else:
                    dx *= 3
            pos = self._slider.mapToParent(rect.center())
            pos += QPoint(int(dx + self.label_shift_x), int(dy + self.label_shift_y))
            if last_edge is not None:
                # prevent label overlap
                if horizontal:
                    pos.setX(int(max(pos.x(), last_edge.x() + label.width() / 2 + 12)))
                else:
                    pos.setY(int(min(pos.y(), last_edge.y() - label.height() / 2 - 4)))
            label.move(pos)
            last_edge = pos
            # label.clearFocus() # better focus behavior without this
            label.show()
        self.update()

    def changeEvent(self, event: QEvent):
        super().changeEvent(event)
        if event.type() in (
            QEvent.Type.PaletteChange,
            QEvent.Type.EnabledChange,
            QEvent.Type.FontChange,
        ):
            style = (
                "QDoubleSpinBox{"
                + f"color: #{self.palette().color(QPalette.ColorRole.Text).rgba():x}"
                ";background:transparent; border: 0;"
                f" font-size:{self.font().pointSize()}pt"
                "}QDoubleSpinBox:!enabled{color: #"
                + f"{self.palette().color(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text).rgba():x}"
                ";}"
            )
            self._slider.setPalette(self.palette())
            self._min_label.setStyleSheet(style)
            self._max_label.setStyleSheet(style)
            self._max_label._update_size()
            self._min_label._update_size()
            for handle in self._handle_labels:
                handle.setStyleSheet(style)
                handle._update_size()

    def value(self) -> tuple[int, int]:
        return super().value()  # type: ignore

    def setValue(self, value) -> None:
        if not self.block_upper:
            self.setMaximum(value[1])
        return super().setValue(value)


class WrappedLabel(QLabel):
    def __init__(
        self,
        text: str = "",
        framed: bool = False,
        align: Qt.AlignmentFlag = Qt.AlignmentFlag.AlignLeft,
    ):
        super().__init__(text)
        if framed:
            self.setBackgroundRole(QPalette.ColorRole.Base)
            self.setAutoFillBackground(True)
            self.setContentsMargins(5, 3, 5, 3)
        self.setAlignment(align)
        self.setWordWrap(True)
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)

    def set_size(self):
        self.setMinimumHeight(0)
        self.setMinimumHeight(max(self.heightForWidth(self.width()), 1))

    def resizeEvent(self, a0: QResizeEvent):
        self.set_size()
        super().resizeEvent(a0)

    def setText(self, text: str):
        # Add Zero-width space in backslashes for proper word wrapping
        super().setText(text.replace("\\", "\\\u200b"))
        self.set_size()

    def text(self):
        return super().text().replace("\u200b", "")


def key_event_modifier(event: QKeyEvent) -> QKeyEvent | None:
    if event.key() == Qt.Key.Key_W:
        return QKeyEvent(event.type(), Qt.Key.Key_Up, event.modifiers())
    if event.key() == Qt.Key.Key_S:
        return QKeyEvent(event.type(), Qt.Key.Key_Down, event.modifiers())
    if event.key() in (Qt.Key.Key_D, Qt.Key.Key_A, Qt.Key.Key_Left, Qt.Key.Key_Right):
        # These keys would be accepted by QTableWidget
        # but we want them to control the VideoPlayer
        event.ignore()
        return None
    return event



from functools import lru_cache
from pathlib import Path

import cv2
import numpy as np


class VideoPathHolder:
    def __init__(self, video_paths: list[Path] | None = None):
        self.video_loaded = False
        self.reduced_cache = False
        if video_paths:
            self.load_paths(video_paths)
        self.cap: cv2.VideoCapture
        self.current_captured_video_path: Path

    def load_paths(self, video_paths: list[Path]) -> None:
        assert video_paths
        self.single_file = len(video_paths) == 1
        self.interval_dict: dict[Path, tuple[int, int]] = {}
        i = 0

        for video_path in video_paths:
            n_frames = int(
                cv2.VideoCapture(str(video_path)).get(cv2.CAP_PROP_FRAME_COUNT)
            )
            self.interval_dict[video_path] = (i, i + n_frames)
            i += n_frames
        self.cap = cv2.VideoCapture(str(video_paths[0]))
        self.current_captured_video_path = video_paths[0]
        self.frame_large_cache.cache_clear()
        self.frame_small_cache.cache_clear()
        self.video_loaded = True

    def set_cache_mode(self, reduced: bool):
        self.reduced_cache = reduced
        if reduced:
            self.frame_large_cache.cache_clear()
        else:
            self.frame_small_cache.cache_clear()

    def frame(self, frame_number: int, color: bool):
        if self.reduced_cache:
            return self.frame_small_cache(frame_number, color)
        return self.frame_large_cache(frame_number, color)

    # TODO check flake8 warnings
    @lru_cache(128)  # noqa: B019
    def frame_large_cache(self, frame_number: int, color: bool):
        return self.read_frame(frame_number, color)

    @lru_cache(16)  # noqa: B019
    def frame_small_cache(self, frame_number: int, color: bool):
        return self.read_frame(frame_number, color)

    def read_frame(self, frame_number: int, color: bool) -> np.ndarray:
        if not self.video_loaded:
            return np.array([[]])
        for _path, (start, end) in self.interval_dict.items():
            if start <= frame_number < end:
                break
        else:
            raise ValueError(
                f"Frame number {frame_number} not in intervals {self.interval_dict}"
            )
        path = _path

        if path != self.current_captured_video_path:
            self.cap = cv2.VideoCapture(str(path))
            self.current_captured_video_path = path

        frame_number_in_path = frame_number - start

        if frame_number_in_path != int(self.cap.get(cv2.CAP_PROP_POS_FRAMES)):
            self.cap.set(cv2.CAP_PROP_POS_FRAMES, frame_number_in_path)
        ret, img = self.cap.read()
        assert ret, f"Error on frame {frame_number}, {frame_number_in_path} of {path}"

        if color:
            return img  # BGR
        return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)



import logging

import torch
from torch.backends import mps


def get_device() -> torch.device:
    """Returns the current available device for PyTorch"""
    if torch.cuda.is_available():
        device = torch.device("cuda")
        logging.info('Using Cuda backend with "%s"', torch.cuda.get_device_name(device))
        return device
    if mps.is_available():
        logging.info("Using MacOS Metal backend")
        return torch.device("mps")
    logging.warning(
        "[bold red]No graphic device was found available[/], running neural"
        " networks on CPU. This may slow down the training steps.",
        extra={"markup": True},
    )
    return torch.device("cpu")


DEVICE = get_device()


def weights_xavier_init(m):
    if isinstance(m, (torch.nn.Linear, torch.nn.Conv2d)):
        torch.nn.init.xavier_uniform_(m.weight.data)


def fc_weights_reinit(m):
    if isinstance(m, torch.nn.Linear):
        torch.nn.init.xavier_uniform_(m.weight.data)


def normalize(tensor: torch.Tensor):
    """Normalize a tensor image with mean and standard deviation.
    Given mean: ``(M1,...,Mn)`` and std: ``(S1,..,Sn)`` for ``n`` channels, this transform
    will normalize each channel of the input ``torch.*Tensor`` i.e.
    ``input[channel] = (input[channel] - mean[channel]) / std[channel]``
    .. note::
        This transform acts out of place, i.e., it does not mutates the input tensor.
    Args:
        tensor (Tensor): Tensor image of size (C, H, W) to be normalized.
    Returns:
        Tensor: Normalized Tensor image.
    """
    # TODO: This is kind of a batch normalization but not trained. Explore using real BN in idCNN.

    mean = torch.tensor([tensor.mean()])
    std = torch.tensor([tensor.std()])
    return tensor.sub_(mean[:, None, None]).div_(std[:, None, None])



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import logging
import pickle
from contextlib import suppress
from itertools import chain, pairwise, product
from multiprocessing import Pool
from pathlib import Path
from typing import Optional

import h5py
import numpy as np

from . import Blob
from .utils import Episode, clean_attrs, resolve_path, track


class ListOfBlobs:
    """Contains all the instances of the class :class:`~blob.Blob` for all
    frames in the video.

    Notes
    -----
    Only frames in the tracking interval defined by the user can have blobs.
    The frames ouside of such interval will be empty.


    Parameters
    ----------
    blobs_in_video : list
        List of lists of blobs. Each element in the outer list represents
        a frame. Each elemtn in each inner list represents a blob in
        the frame.
    """

    def __init__(self, blobs_in_video: list[list[Blob]]):
        logging.info("Generating ListOfBlobs object")
        self.blobs_in_video = blobs_in_video
        self.blobs_are_connected = False
        self.number_of_individual_fragments: int

    @property
    def all_blobs(self):
        return chain.from_iterable(self.blobs_in_video)

    @property
    def number_of_blobs(self) -> int:
        return sum(map(len, self.blobs_in_video))

    @property
    def number_of_crossing_blobs(self) -> int:
        return sum(blob.is_a_crossing for blob in self.all_blobs)

    @property
    def number_of_frames(self):
        return len(self.blobs_in_video)

    @property
    def max_number_of_blobs_in_one_frame(self):
        return max(map(len, self.blobs_in_video))

    def __len__(self):
        return len(self.blobs_in_video)

    def compute_overlapping_between_subsequent_frames(self):
        """Computes overlapping between blobs in consecutive frames.

        Two blobs in consecutive frames overlap if the intersection of the list
        of pixels of both blobs is not empty.

        See Also
        --------
        :meth:`blob.Blob.overlaps_with`
        """

        logging.info("Connecting list of blobs ")

        if self.blobs_are_connected:
            logging.error("List of blobs is already connected")
            return
        # self.disconnect()

        for blobs, blobs_next in pairwise(
            track(self.blobs_in_video, "Connecting blobs")
        ):
            for blob, blob_next in product(blobs, blobs_next):
                if blob.overlaps_with(blob_next):
                    blob.now_points_to(blob_next)
        self.blobs_are_connected = True

        # clean cached property
        with suppress(AttributeError):
            for blob in self.all_blobs:
                del blob.convexHull

    def save(self, path: Path | str):
        """Saves instance of the class

        Parameters
        ----------
        path_to_save : str, optional
            Path where to save the object, by default None
        """
        path = resolve_path(path)
        logging.info(f"Saving ListOfBlobs at {path}")
        path.parent.mkdir(exist_ok=True)
        self.disconnect()

        for blob in self.all_blobs:
            clean_attrs(blob)

        with open(path, "wb") as file:
            pickle.dump(self, file, protocol=pickle.HIGHEST_PROTOCOL)
        self.reconnect()

    @classmethod
    def load(cls, path: Path | str) -> "ListOfBlobs":
        """Loads an instance of a class saved in a .npy file.

        Parameters
        ----------
        blob_list_file : Path
            path to a saved instance of a ListOfBlobs object

        Returns
        -------
        ListOfBlobs
        """
        path = resolve_path(path)
        logging.info(f"Loading ListOfBlobs from {path}")
        if not path.is_file():
            v4_path = path.with_name(
                path.name.replace("list_of_blobs", "blobs_collection")
            ).with_suffix(".npy")

            if v4_path.is_file():
                list_of_blobs = cls.load_from_v4(v4_path)
            else:
                raise FileNotFoundError(path)
        else:
            with open(path, "rb") as file:
                list_of_blobs: ListOfBlobs = pickle.load(file)
        list_of_blobs.reconnect()
        return list_of_blobs

    @classmethod
    def load_from_v4(cls, path: Path) -> "ListOfBlobs":
        logging.info("Loading from v4 file: %s", path)
        list_of_blobs: "ListOfBlobs" = np.load(path, allow_pickle=True).item()

        for blob in track(
            list_of_blobs.all_blobs, "Updating objects from an old idtracker.ai version"
        ):
            blob.is_an_individual = blob._is_an_individual  # type:ignore
            blob.fragment_identifier = blob._fragment_identifier  # type:ignore
            blob.blob_index = blob._blob_index  # type:ignore
            blob.identity = blob._identity  # type:ignore
            blob.identity_corrected_solving_jumps = (
                blob._identity_corrected_solving_jumps  # type:ignore
            )
            blob.identities_corrected_closing_gaps = (
                blob._identities_corrected_closing_gaps  # type:ignore
            )
        return list_of_blobs

    def disconnect(self):
        if self.blobs_are_connected:
            for blob in self.all_blobs:
                blob.next.clear()

    def reconnect(self):
        if self.blobs_are_connected:
            for blob in self.all_blobs:
                for prev_blob in blob.previous:
                    prev_blob.next.append(blob)

    # TODO: this should be part of crossing detector.
    # TODO: the term identification_image should be changed.
    def set_images_for_identification(
        self,
        episodes: list[Episode],
        id_images_file_paths: list[Path],
        id_image_size: list[int],
        bbox_images_path: Path,
        n_jobs: int,
    ):
        """Computes and saves the images used to classify blobs as crossings
        and individuals and to identify the animals along the video.

        Parameters
        ----------
        episodes_start_end : list
            List of tuples of integers indncating the starting and ending
            frames of each episode.
        id_images_file_paths : list
            List of strings indicating the paths to the files where the
            identification images of each episode are stored.
        id_image_size : tuple
            Tuple indicating the width, height and number of channels of the
            identification images.
        number_of_animals : int
            Number of animals to be tracked as indicated by the user.
        number_of_frames : int
            Number of frames in the video
        video_path : str
            Path to the video file
        height : int
            Height of a video frame considering the resolution reduction
            factor.
        width : int
            Width of a video frame considering the resolution reduction factor.
        """

        inputs = [
            (
                bbox_images_path / f"episode_images_{episode.index}.hdf5",
                id_image_size[0],
                file,
                episode,
                self.blobs_in_video[episode.global_start : episode.global_end],
            )
            for file, episode in zip(id_images_file_paths, episodes)
        ]

        with Pool(n_jobs) as p:
            for blobs_in_episode, episode in track(
                p.imap_unordered(self.set_id_images_per_episode, inputs),
                "Setting images for identification",
                len(inputs),
            ):
                self.blobs_in_video[episode.global_start : episode.global_end] = (
                    blobs_in_episode
                )

    @staticmethod
    def set_id_images_per_episode(
        inputs: tuple[Path, int, Path, Episode, list[list[Blob]]]
    ) -> tuple[list[list[Blob]], Episode]:
        bbox_imgs_path, id_image_size, file_path, episode, blobs_in_episode = inputs

        imgs_to_save = np.empty(
            (sum(map(len, blobs_in_episode)), id_image_size, id_image_size), np.uint8
        )

        for index, blob in enumerate(chain.from_iterable(blobs_in_episode)):
            imgs_to_save[index] = blob.get_image_for_identification(
                id_image_size, bbox_imgs_path
            )
            blob.id_image_index = index
            blob.episode = episode.index

        with h5py.File(file_path, "w") as file:
            file.create_dataset("id_images", data=imgs_to_save)

        return blobs_in_episode, episode

    # TODO: maybe move to crossing detector
    def update_id_image_dataset_with_crossings(self, id_images_file_paths: list[Path]):
        """Adds a array to the identification images files indicating whether
        each image is an individual or a crossing.

        Parameters
        ----------
        video : :class:`idtrackerai.video.Video`
            Video object with information about the video and the tracking
            process.
        """
        logging.info("Updating crossings in identification images files")

        crossings = []
        for path in id_images_file_paths:
            with h5py.File(path, "r") as file:
                crossings.append(np.empty(file["id_images"].shape[0], bool))  # type: ignore

        for blob in self.all_blobs:
            id_image_index = blob.id_image_index

            crossings[blob.episode][id_image_index] = blob.is_a_crossing

        for path, crossing in zip(id_images_file_paths, crossings):
            with h5py.File(path, "r+") as file:
                file.create_dataset("crossings", data=crossing)

    def remove_centroid(self, frame_number: int, centroid_to_remove, id_to_remove):
        for blob in self.blobs_in_video[frame_number]:
            for indx, (id, centroid) in enumerate(
                zip(blob.all_final_identities, blob.all_final_centroids)
            ):
                if id == id_to_remove:
                    dist = (centroid[0] - centroid_to_remove[0]) ** 2 + (
                        centroid[1] - centroid_to_remove[1]
                    ) ** 2
                    if dist < 1:  # it is the same centroid
                        blob.init_validator_variables()
                        blob.user_generated_centroids[indx] = (-1, -1)
                        blob.user_generated_identities[indx] = -1

    def reset_user_generated_corrections(
        self, start_frame: int = 0, end_frame: Optional[int] = None
    ):
        """[Validation] Resets the identities and centroids generetad by the user.

        Parameters
        ----------
        start_frame : int
            Frame from which to start resetting identities and centroids
        end_frame : int
            Frame where to end resetting identities and centroids
        """

        for blobs_in_frame in track(
            self.blobs_in_video[start_frame:end_frame], "Resetting user corrections"
        ):
            # Reset all user generated identities and centroids
            for blob in blobs_in_frame:
                if blob.added_by_user:
                    self.blobs_in_video[blob.frame_number].remove(blob)
                else:
                    blob.user_generated_identities = None
                    blob.user_generated_centroids = None

    def update_centroid(
        self, frame_number: int, centroid_id: int, old_centroid, new_centroid
    ):
        old_centroid = tuple(old_centroid)
        new_centroid = tuple(new_centroid)
        blobs_in_frame = self.blobs_in_video[frame_number]
        assert blobs_in_frame

        dist_to_old_centroid: list[tuple[Blob, float]] = []

        for blob in blobs_in_frame:
            try:
                indx, centroid, dist = blob.index_and_centroid_closer_to(
                    old_centroid, centroid_id
                )
            except ValueError:  # blob has not centroid_id
                pass
            else:
                dist_to_old_centroid.append((blob, dist))

        blob_with_old_centroid = min(dist_to_old_centroid, key=lambda x: x[1])[0]
        blob_with_old_centroid.update_centroid(old_centroid, new_centroid, centroid_id)

    def add_centroid(self, frame_number: int, identity: int, centroid):
        centroid = tuple(centroid)
        blobs_in_frame = self.blobs_in_video[frame_number]
        if not blobs_in_frame:
            self.add_blob(frame_number, centroid, identity)
            return

        for blob in blobs_in_frame:
            if blob.contains_point(centroid):
                blob.add_centroid(centroid, identity)
                return

        blob = min(blobs_in_frame, key=lambda b: b.distance_from_countour_to(centroid))
        blob.add_centroid(centroid, identity)

    def add_blob(self, frame_number: int, centroid: tuple, identity: int):
        """[Validation] Adds a Blob object the frame number.

        Adds a Blob object to a given frame_number with a given centroid and
        identity. Note that this Blob won't have most of the features (e.g.
        area, contour, fragment_identifier, bbox, ...). It is only
        intended to be used for validation and correction of trajectories.
        The new blobs generated are considered to be individuals.

        Parameters
        ----------
        frame_number : int
            Frame in which the new blob will be added
        centroid : tuple
            The centroid of the new blob
        identity : int
            Identity of the new blob
        """
        contour = np.array(
            [
                [centroid[0] - 2, centroid[1] - 2],
                [centroid[0] - 2, centroid[1] + 2],
                [centroid[0] + 2, centroid[1] + 2],
                [centroid[0] + 2, centroid[1] - 2],
            ],
            int,
        )
        new_blob = Blob(contour, frame_number)
        new_blob.added_by_user = True
        new_blob.user_generated_centroids = [(centroid[0], centroid[1])]
        new_blob.user_generated_identities = [identity]
        new_blob.is_an_individual = True
        self.blobs_in_video[frame_number].append(new_blob)

    @property
    def maximum_number_of_blobs(self):
        return max(map(len, self.blobs_in_video))



# This file is part of idtracker.ai a multiple animals tracking system
# described in [1].
# Copyright (C) 2017- Francisco Romero Ferrero, Mattia G. Bergomi,
# Francisco J.H. Heras, Robert Hinz, Gonzalo G. de Polavieja and the
# Champalimaud Foundation.
#
# idtracker.ai is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details. In addition, we require
# derivatives or applications to acknowledge the authors by citing [1].
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For more information please send an email (idtrackerai@gmail.com) or
# use the tools available at https://gitlab.com/polavieja_lab/idtrackerai.git.
#
# [1] Romero-Ferrero, F., Bergomi, M.G., Hinz, R.C., Heras, F.J.H.,
# de Polavieja, G.G., Nature Methods, 2019.
# idtracker.ai: tracking all individuals in small or large collectives of
# unmarked animals.
# (F.R.-F. and M.G.B. contributed equally to this work.
# Correspondence should be addressed to G.G.d.P:
# gonzalo.polavieja@neuro.fchampalimaud.org)
import json
import logging
import pickle
from itertools import combinations
from math import comb
from pathlib import Path
from typing import Any, Iterable, Literal

import h5py
import numpy as np

from . import Blob, Fragment, GlobalFragment
from .utils import clean_attrs, load_id_images, resolve_path, track


class ListOfFragments:
    """Contains all the instances of the class :class:`fragment.Fragment`.

    Parameters
    ----------
    fragments : list
        List of instances of the class :class:`fragment.Fragment`.
    id_images_file_paths : list
        List of strings with the paths to the files where the identification
        images are stored.
    """

    accumulable_individual_fragments: set[int]
    not_accumulable_individual_fragments: set[int]

    def __init__(
        self,
        fragments: list[Fragment],
        id_images_file_paths: list[Path],
        number_of_animals: int,
    ):
        # Assert fragments are sorted
        for i, fragment in enumerate(fragments):
            assert i == fragment.identifier
        self.n_animals = number_of_animals
        self.fragments = fragments
        self.id_images_file_paths = id_images_file_paths
        self.connect_coexisting_fragments()

    @property
    def number_of_fragments(self):
        return len(self.fragments)

    @property
    def individual_fragments(self):
        return (frag for frag in self.fragments if frag.is_an_individual)

    # TODO: if the resume feature is not active, this does not make sense|
    def reset(self, roll_back_to: Literal["fragmentation", "accumulation"]):
        """Resets all the fragment to a given processing step.

        Parameters
        ----------
        roll_back_to : str
            Name of the step at which the fragments should be reset.
            It can be 'fragmentation', 'pretraining', 'accumulation' or
            'assignment'

        See Also
        --------
        :meth:`fragment.Fragment.reset`
        """
        logging.info(f"Resetting ListOfFragments to '{roll_back_to}'", stacklevel=3)
        for fragment in self.fragments:
            fragment.reset(roll_back_to, self.n_animals)

    # TODO: maybe this should go to the accumulator manager
    def get_images_from_fragments_to_assign(self):
        """Take all the fragments that have not been used to train the idCNN
        and that are associated with an individual, and concatenates their
        images in order to feed them to the identification network.

        Returns
        -------
        ndarray
            [number_of_images, height, width, number_of_channels]
        """
        images: list[tuple[int, int]] = []
        for fragment in self.individual_fragments:
            if not fragment.used_for_training:
                images += fragment.image_locations

        logging.info(
            f"Number of images to identify non-accumulated fragments: {len(images)}"
        )
        return load_id_images(self.id_images_file_paths, images)

    # TODO: The following methods depend on the identification strategy.

    @property
    def number_of_images_in_global_fragments(self) -> int:
        """Number of images available in global fragments
        (without repetitions)"""
        return sum(
            fragment.number_of_images
            for fragment in self.fragments
            if fragment.identifier in self.accumulable_individual_fragments
        )

    @property
    def ratio_of_images_used_for_pretraining(self) -> float:
        """Returns the ratio of images used for pretraining over the number of
        available images"""
        return (
            sum(
                fragment.number_of_images
                for fragment in self.fragments
                if fragment.used_for_pretraining
            )
            / self.number_of_images_in_global_fragments
        )

    @property
    def ratio_of_images_used_for_training(self) -> float:
        """Returns the ratio of images used for training over the number of
        available images"""
        return (
            sum(
                fragment.number_of_images
                for fragment in self.fragments
                if fragment.used_for_training
            )
            / self.number_of_images_in_global_fragments
        )

    def compute_P2_vectors(self):
        """Computes the P2_vector associated to every individual fragment. See
        :meth:`fragment.Fragment.compute_P2_vector`
        """
        for fragment in self.individual_fragments:
            fragment.compute_P2_vector(self.n_animals)

    def get_number_of_unidentified_individual_fragments(self):
        """Returns the number of individual fragments that have not been
        identified during the fingerprint protocols cascade

        Returns
        -------
        int
            number of non-identified individual fragments
        """
        return sum(
            frag.is_an_individual and not frag.used_for_training
            for frag in self.fragments
        )

    def get_next_fragment_to_identify(self) -> Fragment | None:
        """Returns the next fragment to be identified after the cascade of
        training and identitication protocols by sorting according to the
        certainty computed with P2. See :attr:fragment.Fragment.certainty_P2`

        Returns
        -------
        :class:`fragment.Fragment`
            An instance of the class :class:`fragment.Fragment`
        """
        try:
            return max(
                filter(
                    lambda frag: frag.is_an_individual
                    and frag.assigned_identities[0] is None,
                    self.fragments,
                ),
                key=lambda frag: frag.certainty_P2,
            )
        except ValueError:
            return None

    def update_id_images_dataset(self):
        """Updates the identification images files with the identity assigned
        to each fragment during the tracking process.
        """
        logging.info("Updating identities in identification images files")

        identities = []
        for path in self.id_images_file_paths:
            with h5py.File(path, "r") as file:
                identities.append(np.full(file["id_images"].shape[0], 0))  # type: ignore

        for fragment in self.fragments:
            if fragment.used_for_training:
                for image, episode in fragment.image_locations:
                    identities[episode][image] = fragment.identity

        for path, identities_in_episode in zip(self.id_images_file_paths, identities):
            with h5py.File(path, "r+") as file:
                dataset = file.require_dataset(
                    "identities", shape=len(identities_in_episode), dtype=int
                )
                dataset[:] = identities_in_episode

    def get_ordered_list_of_fragments(
        self, scope: Literal["to_the_past", "to_the_future"], specific_frame: int
    ) -> list[Fragment]:
        """Sorts the fragments starting from the frame number
        `first_frame_first_global_fragment`. According to `scope` the sorting
        is done either "to the future" of "to the past" with respect to
        `first_frame_first_global_fragment`

        Parameters
        ----------
        scope : str
            either "to_the_past" or "to_the_future"
        first_frame_first_global_fragment : int
            frame number corresponding to the first frame in which all the
            individual fragments coexist in the first global fragment using
            in an iteration of the fingerprint protocol cascade

        Returns
        -------
        list
            list of sorted fragments

        """
        if scope == "to_the_past":
            fragments_to_the_past = filter(
                lambda frag: frag.end_frame <= specific_frame, self.fragments
            )
            return sorted(
                fragments_to_the_past, key=lambda x: x.end_frame, reverse=True
            )
        if scope == "to_the_future":
            fragments_to_the_future = filter(
                lambda frag: frag.start_frame >= specific_frame, self.fragments
            )
            return sorted(fragments_to_the_future, key=lambda x: x.start_frame)
        raise ValueError(scope)

    def save(self, path: Path | str):
        """Save an instance of the object in disk,

        Parameters
        ----------
        fragments_path : str
            Path where the instance of the object will be stored.
        """
        path = resolve_path(path)
        logging.info(f"Saving ListOfFragments as {path}")
        path.parent.mkdir(exist_ok=True)

        json.dump(self, path.open("w"), cls=FragmentsEncoder, indent=4)

    @classmethod
    def load(cls, path: Path | str, reconnect=True) -> "ListOfFragments":
        """Loads a previously saved (see :meth:`save`) from the path
        `path_to_load`
        """
        path = resolve_path(path)
        logging.info(f"Loading ListOfFragments from {path}")

        if not path.is_file() and path.with_suffix(".pickle").is_file():
            # <=5.1.3 compatibility
            pickle.load(path.with_suffix(".pickle").open("rb")).save(path)

        list_of_fragments = cls.__new__(cls)
        json_data = json.load(path.with_suffix(".json").open("r"))

        list_of_fragments.accumulable_individual_fragments = set(
            json_data.get("accumulable_individual_fragments", [])
        )
        list_of_fragments.not_accumulable_individual_fragments = set(
            json_data.get("not_accumulable_individual_fragments", [])
        )
        if "number_of_animals" in json_data:
            list_of_fragments.n_animals = json_data["number_of_animals"]
        if "n_animals" in json_data:
            list_of_fragments.n_animals = json_data["n_animals"]

        list_of_fragments.id_images_file_paths = list(
            map(Path, json_data["id_images_file_paths"])
        )

        list_of_fragments.fragments = [
            Fragment.from_json(frag_data) for frag_data in json_data["fragments"]
        ]

        for fragment in list_of_fragments.fragments:
            if (
                fragment.identifier
                in list_of_fragments.accumulable_individual_fragments
            ):
                fragment.accumulable = True
            elif (
                fragment.identifier
                in list_of_fragments.not_accumulable_individual_fragments
            ):
                fragment.accumulable = False

        if reconnect:
            list_of_fragments.connect_coexisting_fragments()

        return list_of_fragments

    def connect_coexisting_fragments(self):
        logging.info("Connecting coexisting individual fragments")
        # Make it N (not N²) with, maybe, sets (not lists)
        for fragment in self.fragments:
            fragment.coexisting_individual_fragments = []

        for fragment_A, fragment_B in track(
            combinations(self.fragments, 2),
            "Connecting coexisting fragments",
            comb(len(self.fragments), 2),
        ):
            if fragment_A.coexist_with(fragment_B):
                if fragment_A.is_an_individual:
                    fragment_B.coexisting_individual_fragments.append(fragment_A)
                if fragment_B.is_an_individual:
                    fragment_A.coexisting_individual_fragments.append(fragment_B)

    def manage_accumulable_non_accumulable_fragments(
        self,
        accumulable_global_fragments: list[GlobalFragment],
        non_accumulable_global_fragments: list[GlobalFragment],
    ):
        """Gets the unique identifiers associated to individual fragments that
        can be accumulated.

        Parameters
        ----------
        list_of_global_fragments : :class:`list_of_global_fragments.ListOfGlobalFragments`
            Object collecting the global fragment objects (instances of the
            class :class:`globalfragment.GlobalFragment`) detected in the
            entire video.

        """
        self.accumulable_individual_fragments = {
            identifier
            for glob_frag in accumulable_global_fragments
            for identifier in glob_frag.individual_fragments_identifiers
        }
        self.not_accumulable_individual_fragments = {
            identifier
            for glob_frag in non_accumulable_global_fragments
            for identifier in glob_frag.individual_fragments_identifiers
        } - self.accumulable_individual_fragments

        for fragment in self.fragments:
            if fragment.identifier in self.accumulable_individual_fragments:
                fragment.accumulable = True
            elif fragment.identifier in self.not_accumulable_individual_fragments:
                fragment.accumulable = False

    @property
    def number_of_crossing_fragments(self) -> int:
        return sum(fragment.is_a_crossing for fragment in self.fragments)

    @property
    def number_of_individual_fragments(self) -> int:
        return sum(1 for _ in self.individual_fragments)

    @property
    def number_of_individual_fragments_not_in_a_glob_fragment(self) -> int:
        return sum(
            not fragment.is_in_a_global_fragment
            for fragment in self.individual_fragments
        )

    @property
    def number_of_accumulable_individual_fragments(self) -> int:
        return len(self.accumulable_individual_fragments)

    @property
    def number_of_not_accumulable_individual_fragments(self) -> int:
        return len(self.not_accumulable_individual_fragments)

    @property
    def number_of_blobs(self) -> int:
        return sum(fragment.number_of_images for fragment in self.fragments)

    @property
    def number_of_crossing_blobs(self) -> int:
        return sum(
            fragment.is_a_crossing * fragment.number_of_images
            for fragment in self.fragments
        )

    @property
    def number_of_individual_blobs(self) -> int:
        return sum(fragment.number_of_images for fragment in self.individual_fragments)

    @property
    def number_of_individual_blobs_not_in_a_global_fragment(self) -> int:
        return sum(
            not fragment.is_in_a_global_fragment * fragment.number_of_images
            for fragment in self.individual_fragments
        )

    @property
    def fragments_not_accumulated(self) -> set[int]:
        return self.accumulable_individual_fragments & {
            fragment.identifier
            for fragment in self.fragments
            if not fragment.used_for_training
        }

    @property
    def number_of_globally_accumulated_individual_fragments(self) -> int:
        return sum(
            fragment.accumulated_globally for fragment in self.individual_fragments
        )

    @property
    def number_of_partially_accumulated_individual_fragments(self) -> int:
        return sum(
            fragment.accumulated_partially for fragment in self.individual_fragments
        )

    @property
    def number_of_accumulable_individual_blobs(self) -> int:
        return sum(
            bool(fragment.accumulable) * fragment.number_of_images
            for fragment in self.fragments
        )

    @property
    def number_of_not_accumulable_individual_blobs(self) -> int:
        return sum(
            (not fragment.accumulable) * fragment.number_of_images
            for fragment in self.fragments
            if fragment.accumulable is not None
        )

    @property
    def number_of_globally_accumulated_individual_blobs(self) -> int:
        return sum(
            fragment.accumulated_globally * fragment.number_of_images
            for fragment in self.individual_fragments
        )

    @property
    def number_of_partially_accumulated_individual_blobs(self) -> int:
        return sum(
            fragment.accumulated_partially * fragment.number_of_images
            for fragment in self.individual_fragments
        )

    def get_stats(self) -> dict[str, Any]:
        """Collects the following counters from the fragments.

        * number_of_fragments
        * number_of_crossing_fragments
        * number_of_individual_fragments
        * number_of_individual_fragments_not_in_a_glob_fragment
        * number_of_accumulable_individual_fragments
        * number_of_not_accumulable_individual_fragments
        * number_of_accumulated_individual_fragments
        * number_of_globally_accumulated_individual_fragments
        * number_of_partially_accumulated_individual_fragments
        * number_of_blobs
        * number_of_crossing_blobs
        * number_of_individual_blobs
        * number_of_individual_blobs_not_in_a_global_fragment
        * number_of_accumulable_individual_blobs
        * number_of_not_accumulable_individual_blobs
        * number_of_accumulated_individual_blobs
        * number_of_globally_accumulated_individual_blobs
        * number_of_partially_accumulated_individual_blobs

        Returns
        -------
        dict
            Dictionary with the counters mentioned above

        """

        stats: dict[str, Any] = {
            "fragments": self.number_of_fragments,
            "crossing_fragments": self.number_of_crossing_fragments,
            "individual_fragments": self.number_of_individual_fragments,
            "individual_fragments_not_in_a_global_fragment": (
                self.number_of_individual_fragments_not_in_a_glob_fragment
            ),
            "accumulable_individual_fragments": (
                self.number_of_accumulable_individual_fragments
            ),
            "not_accumulable_individual_fragments": (
                self.number_of_not_accumulable_individual_fragments
            ),
            "globally_accumulated_individual_fragments": (
                self.number_of_globally_accumulated_individual_fragments
            ),
            "partially_accumulated_individual_fragments": (
                self.number_of_partially_accumulated_individual_fragments
            ),
            "blobs": self.number_of_blobs,
            "crossing_blobs": self.number_of_crossing_blobs,
            "individual_blobs": self.number_of_individual_blobs,
            "individual_blobs_not_in_a_global_fragment": (
                self.number_of_individual_blobs_not_in_a_global_fragment
            ),
            "accumulable_individual_blobs": self.number_of_accumulable_individual_blobs,
            "not_accumulable_individual_blobs": (
                self.number_of_not_accumulable_individual_blobs
            ),
            "globally_accumulated_individual_blobs": (
                self.number_of_globally_accumulated_individual_blobs
            ),
            "partially_accumulated_individual_blobs": (
                self.number_of_partially_accumulated_individual_blobs
            ),
        }

        log = "Final statistics:"
        for key, value in stats.items():
            log += f"\n  {value} {key.replace('_', ' ')}"
        logging.info(log)

        return stats

    @classmethod
    def from_fragmented_blobs(
        cls,
        all_blobs: Iterable[Blob],
        number_of_animals: int,
        id_images_file_paths: list[Path],
    ) -> "ListOfFragments":
        """Generate a list of instances of :class:`fragment.Fragment` collecting
        all the fragments in the video.

        Parameters
        ----------
        blobs_in_video : list
            list of the blob objects (see class :class:`blob.Blob`) generated
            from the blobs segmented in the video
        number_of_animals : int
            Number of animals to track as defined by the user

        Returns
        -------
        list
            list of instances of :class:`fragment.Fragment`

        """
        fragments: list[Fragment] = []
        used_fragment_identifiers: set[int] = set()

        logging.info("Creating list of fragments")
        for blob in all_blobs:
            current_fragment_identifier = blob.fragment_identifier
            if current_fragment_identifier not in used_fragment_identifiers:
                images = [blob.id_image_index]
                centroids = [blob.centroid]
                episodes = [blob.episode]
                start = blob.frame_number
                current = blob

                while (
                    current.n_next > 0
                    and current.next[0].fragment_identifier
                    == current_fragment_identifier
                ):
                    current = current.next[0]
                    images.append(current.id_image_index)
                    centroids.append(current.centroid)
                    episodes.append(current.episode)

                end = current.frame_number

                fragment = Fragment(
                    current_fragment_identifier,
                    start,
                    end + 1,  # it is not inclusive
                    images,
                    centroids,
                    episodes,
                    blob.is_an_individual,
                )
                used_fragment_identifiers.add(current_fragment_identifier)
                fragments.append(fragment)
        return cls(fragments, id_images_file_paths, number_of_animals)

    def update_blobs(self, all_blobs: Iterable[Blob]):
        """Updates the blobs objects generated from the video with the
        attributes computed for each fragment

        Parameters
        ----------
        fragments : list
            List of all the fragments

        See Also
        --------
        :meth:`blob.Blob.compute_fragment_identifier_and_blob_index`

        """
        logging.info("Updating list of blobs from list of fragments")
        for blob in all_blobs:
            fragment = self.fragments[blob.fragment_identifier]
            blob.identity = fragment.identity
            blob.identity_corrected_solving_jumps = (
                fragment.identity_corrected_solving_jumps
            )
            blob.user_generated_identity = fragment.user_generated_identity
            blob.is_an_individual = fragment.is_an_individual
            if fragment.forced_crossing:
                blob.forced_crossing = True


class FragmentsEncoder(json.JSONEncoder):
    def default(self, obj):
        match obj:
            case Path():
                return str(obj)

            case ListOfFragments():
                serial = obj.__dict__.copy()
                serial["accumulable_individual_fragments"] = (
                    f"NotString{json.dumps(list(serial.get('accumulable_individual_fragments',{})))}"
                )
                serial["not_accumulable_individual_fragments"] = (
                    f"NotString{json.dumps(list(serial.get('not_accumulable_individual_fragments',{})))}"
                )
                return serial

            case Fragment():
                clean_attrs(obj)
                serial = obj.__dict__.copy()
                serial.pop("coexisting_individual_fragments", None)
                serial.pop("centroids", None)  # v5.1.3 compatibility
                serial.pop("accumulable", None)
                serial["images"] = "NotString" + json.dumps(obj.images)
                if len(set(obj.episodes)) == 1:
                    # compress when all images are in the same episode
                    serial["episodes"] = f"NotString{[obj.episodes[0]]}"
                else:
                    serial["episodes"] = "NotString" + json.dumps(obj.episodes)
                if "frame_by_frame_velocity" in serial:
                    serial["frame_by_frame_velocity"] = "NotString" + json.dumps(
                        np.round(obj.frame_by_frame_velocity, 2).tolist()
                    )
                if "start_position" in serial:
                    serial["start_position"] = "NotString" + json.dumps(
                        np.round(obj.start_position, 2).tolist()
                    )
                if "end_position" in serial:
                    serial["end_position"] = "NotString" + json.dumps(
                        np.round(obj.end_position, 2).tolist()
                    )
                for key in ("P1_vector", "P2_vector", "ambiguous_identities"):
                    if key in serial:
                        serial[key] = "NotString" + json.dumps(serial[key].tolist())

                return serial
            case np.integer():
                return int(obj)
            case np.bool_():
                return bool(obj)
            case np.floating():
                return float(obj)
            case _:
                return super().default(obj)

    def iterencode(self, obj, **kwargs):
        for encoded in super().iterencode(obj, **kwargs):
            if encoded.startswith('"NotString'):
                # remove first and final '"NoIndent..."' and remove indents,
                yield encoded[10:-1]
            else:
                yield encoded



import logging
import random
from pathlib import Path
from typing import Literal

import numpy as np
from torch.nn import Module

from idtrackerai import Fragment, GlobalFragment, ListOfFragments, ListOfGlobalFragments
from idtrackerai.network import NetworkParams
from idtrackerai.utils import conf, load_id_images

from .identity_network import get_predictions_identities


class AccumulationManager:
    """Manages the process of accumulating images for training the network.

    Attributes
    ----------

    list_of_global_fragments: ListOfGlobalFragments
        Collection of global fragments
    counter : int
        Number of iterations for an instantiation
    certainty_threshold: float
        Value in [0,1] to establish if the identitification of a fragment
        is certain.
    threshold_acceptable_accumulation: float
        Value in [0,1] to establish if an accumulation is acceptable
    accumulation_strategy: string
        Accepts "global" and "partial" in order to perform either partial or
        global accumulation.
    used_images : nd.array
        images used for training the network
    used_labels : nd.array
        labels for the images used for training
    new_images : nd.array
        set of images that will be added to the new training
    new_labels : nd.array
        labels for the set of images that will be added for training
    """

    def __init__(
        self,
        id_images_file_paths: list[Path],
        number_of_animals: int,
        list_of_fragments: ListOfFragments,
        list_of_global_fragments: ListOfGlobalFragments,
        certainty_threshold: float | None = None,
        threshold_acceptable_accumulation: float | None = None,
    ):
        logging.info("Initializing accumulation manager")
        if certainty_threshold is None:
            certainty_threshold = float(conf.CERTAINTY_THRESHOLD)
        if threshold_acceptable_accumulation is None:
            threshold_acceptable_accumulation = float(
                conf.THRESHOLD_ACCEPTABLE_ACCUMULATION
            )
        self.id_images_file_paths = id_images_file_paths
        self.n_animals = number_of_animals
        self.list_of_fragments = list_of_fragments
        self.list_of_global_fragments = list_of_global_fragments
        self.current_step: int = 0
        self.certainty_threshold = certainty_threshold
        self.threshold_acceptable_accumulation = threshold_acceptable_accumulation
        self.accumulation_strategy: Literal["global", "partial"] = "global"
        self.individual_fragments_used: set[int] = set()
        """set with the individual_fragments_identifiers of the individual
        fragments used for training"""
        self.temporary_individual_fragments_used: set[int] = set()

        self.used_images = None
        self.used_labels = None
        self.new_images = None
        self.new_labels = None
        self.ratio_accumulated_images: float
        # When we init the Accumulation manager we are starting Protocol 1
        # or the accumulation parachute (
        self.threshold_early_stop_accumulation: float = (
            conf.THRESHOLD_EARLY_STOP_ACCUMULATION
        )

    @property
    def new_global_fragments_for_training(self) -> bool:
        """We stop the accumulation when there are not more global fragments
        that are acceptable for training."""
        there_are = any(
            (
                global_fragment.acceptable_for_training(self.accumulation_strategy)
                and not global_fragment.used_for_training
            )
            for global_fragment in self.list_of_global_fragments.global_fragments
        )

        logging.info(
            (
                "[bold]There are global fragments acceptable for training"
                if there_are
                else "[bold]There are no more global fragments acceptable for training"
            ),
            extra={"markup": True},
        )

        return there_are

    def get_new_images_and_labels(self):
        """Get the images and labels of the new global fragments that are going
        to be used for training. This function checks whether the images of a individual
        fragment have been added before"""

        images = []
        labels = []
        for fragment in self.list_of_fragments.individual_fragments:
            if fragment.acceptable_for_training and not fragment.used_for_training:
                images += fragment.image_locations
                labels += [fragment.temporary_id] * fragment.number_of_images

        if images:
            self.new_images, self.new_labels = np.asarray(images), np.asarray(labels)
        else:
            self.new_images, self.new_labels = None, None

        n_used_images = len(self.used_images) if self.used_images is not None else 0
        n_new_images = len(self.new_images) if self.new_images is not None else 0
        n_images = n_used_images + n_new_images

        if n_new_images:
            logging.info("%d new images for training", n_new_images)
        else:
            logging.info("There are no new images in this accumulation")

        if n_used_images:
            logging.info("%d old images for training", n_used_images)

        ratio = n_images / self.list_of_fragments.number_of_images_in_global_fragments
        logging.info(
            f"{n_images} images in total, {ratio:.2%} of the total accumulable"
        )

    def get_images_and_labels_for_training(self):
        """Create a new dataset of labelled images to train the idCNN in the
        following way:
        Per individual select conf.MAXIMAL_IMAGES_PER_ANIMAL images.
        Such collection of images is composed
        of a ratio corresponding to conf.RATIO_NEW of new images (acquired in
        the current evaluation of the
        global fragments) and conf.RATIO_OLD of images already used
        in the previous iteration."""
        logging.info("Getting images for training...")
        random.seed(0)
        images = []
        labels = []
        for i in range(self.n_animals):
            if self.new_labels is None:
                new_images_indices = np.asarray([], int)
            else:
                new_images_indices = np.argwhere(self.new_labels == i)[:, 0]

            if self.used_labels is None:
                used_images_indices = np.asarray([], int)
            else:
                used_images_indices = np.argwhere(self.used_labels == i)[:, 0]
            number_of_new_images = len(new_images_indices)
            number_of_used_images = len(used_images_indices)
            number_of_images_for_individual = (
                number_of_new_images + number_of_used_images
            )
            if number_of_images_for_individual > conf.MAXIMAL_IMAGES_PER_ANIMAL:
                # we take a proportion of the old images a new images only if the
                # total number of images for this label is bigger than the
                # limit conf.MAXIMAL_IMAGES_PER_ANIMAL
                number_samples_new = int(
                    conf.MAXIMAL_IMAGES_PER_ANIMAL * conf.RATIO_NEW
                )
                number_samples_used = (
                    conf.MAXIMAL_IMAGES_PER_ANIMAL - number_samples_new
                )
                if number_of_used_images < number_samples_used:
                    # if the proportion of used images is bigger than the number of
                    # used images we take all the used images for this label and update
                    # the number of new images to reach the conf.MAXIMAL_IMAGES_PER_ANIMAL
                    number_samples_used = number_of_used_images
                    number_samples_new = (
                        conf.MAXIMAL_IMAGES_PER_ANIMAL - number_samples_used
                    )
                if number_of_new_images < number_samples_new:
                    # if the proportion of new images is bigger than the number of
                    # new images we take all the new images for this label and update
                    # the number of used images to reac the conf.MAXIMAL_IMAGES_PER_ANIMAL
                    number_samples_new = number_of_new_images
                    number_samples_used = (
                        conf.MAXIMAL_IMAGES_PER_ANIMAL - number_samples_new
                    )
                # we put together a random sample of the new images and the used images
                if self.new_images is not None:
                    images += random.sample(
                        list(self.new_images[new_images_indices]), number_samples_new
                    )
                    labels += [i] * number_samples_new
                if self.used_images is not None:
                    # this condition is set because the first time we accumulate
                    # the variable used_images is None
                    images += random.sample(
                        list(self.used_images[used_images_indices]), number_samples_used
                    )
                    labels += [i] * number_samples_used
            else:
                # if the total number of images for this label does not exceed
                # the conf.MAXIMAL_IMAGES_PER_ANIMAL
                # we take all the new images and all the used images
                if self.new_images is not None:
                    images += list(self.new_images[new_images_indices])
                    labels += [i] * number_of_new_images
                if self.used_images is not None:
                    # this condition is set because the first time we accumulate
                    # the variable used_images is None
                    images += list(self.used_images[used_images_indices])
                    labels += [i] * number_of_used_images
        return load_id_images(self.id_images_file_paths, images), np.asarray(labels)

    def update_used_images_and_labels(self):
        """Sets as used the images already used for training"""
        logging.info("Update images and labels used for training")
        if self.current_step == 0:
            self.used_images = self.new_images
            self.used_labels = self.new_labels
        elif self.new_images is not None:
            self.used_images = np.concatenate(
                (self.used_images, self.new_images), axis=0
            )
            self.used_labels = np.concatenate(
                [self.used_labels, self.new_labels], axis=0
            )

    def update_fragments_used_for_training(self):
        """Once a global fragment has been used for training, sets the flags
        used_for_training to TRUE and acceptable_for_training to FALSE"""
        logging.info("Updating fragments used for training")
        for fragment in self.list_of_fragments.fragments:
            if fragment.acceptable_for_training and not fragment.used_for_training:
                fragment.used_for_training = True
                fragment.acceptable_for_training = False
                fragment.set_partially_or_globally_accumulated(
                    self.accumulation_strategy
                )
                fragment.accumulation_step = self.current_step

    def assign_identities_to_fragments_used_for_training(self):
        """Assign the identities to the global fragments used for training and
        their individual fragments.
        This function checks that the identities of the individual fragments in
        the global fragment
        are consistent with the previously assigned identities
        """
        logging.info("Assigning identities to accumulated global fragments")
        for fragment in self.list_of_fragments.fragments:
            if fragment.used_for_training:
                assert fragment.temporary_id is not None
                fragment.identity = fragment.temporary_id + 1
                fragment.P1_vector[:] = 0.0
                fragment.P1_vector[fragment.temporary_id] = 1.0

    def update_set_of_individual_fragments_used(self):
        """Updates the list of individual fragments used for training and
        their identities.
        If an individual fragment was added before is not added again.
        """
        logging.info("Updating list of individual fragments used for training")
        self.individual_fragments_used = {
            fragment.identifier
            for fragment in self.list_of_fragments.fragments
            if fragment.used_for_training
        }

    def split_predictions_after_network_assignment(
        self,
        predictions,
        softmax_probs,
        indices_to_split,
        candidate_individual_fragments_identifiers: list[int],
    ):
        """Gathers predictions relative to fragment images from the GPU and
        splits them according to their organization in fragments.
        """
        logging.debug("Computing fragment prediction statistics")
        individual_fragments_predictions = np.split(predictions, indices_to_split)
        individual_fragments_softmax_probs = np.split(softmax_probs, indices_to_split)

        for (
            individual_fragment_predictions,
            individual_fragment_softmax_probs,
            candidate_individual_fragment_identifier,
        ) in zip(
            individual_fragments_predictions,
            individual_fragments_softmax_probs,
            candidate_individual_fragments_identifiers,
        ):
            self.list_of_fragments.fragments[
                candidate_individual_fragment_identifier
            ].compute_identification_statistics(
                individual_fragment_predictions,
                individual_fragment_softmax_probs,
                self.list_of_fragments.n_animals,
            )

    def reset_accumulation_variables(self):
        """After an accumulation is finished reinitialise the variables involved
        in the process.
        """
        self.temporary_individual_fragments_used.clear()
        if self.accumulation_strategy == "global":
            self.number_of_noncertain_global_fragments = 0
            self.number_of_random_assigned_global_fragments = 0
            self.number_of_nonconsistent_global_fragments = 0
            self.number_of_nonunique_global_fragments = 0
        self.number_of_sparse_fragments = 0
        self.number_of_noncertain_fragments = 0
        self.number_of_random_assigned_fragments = 0
        self.number_of_nonconsistent_fragments = 0
        self.number_of_nonunique_fragments = 0
        self.number_of_acceptable_fragments = 0

    def print_accumulation_variables(self):
        lines = (
            "Prediction results:",
            (
                "Non certain global fragments:"
                f" {self.number_of_noncertain_global_fragments}"
            ),
            (
                "Randomly assigned global fragments:"
                f" {self.number_of_random_assigned_global_fragments}"
            ),
            (
                "Non consistent global fragments:"
                f" {self.number_of_nonconsistent_global_fragments}"
            ),
            f"Non unique global fragments: {self.number_of_nonunique_global_fragments}",
            (
                "Acceptable global fragments:"
                f" {self.number_of_acceptable_global_fragments}"
            ),
            f"Non certain fragments: {self.number_of_noncertain_fragments}",
            f"Randomly assigned fragments: {self.number_of_random_assigned_fragments}",
            f"Non consistent fragments: {self.number_of_nonconsistent_fragments}",
            f"Non unique fragments: {self.number_of_nonunique_fragments}",
            f"Acceptable fragments: {self.number_of_acceptable_fragments}",
        )
        logging.info("\n    ".join(lines))

    def get_acceptable_global_fragments_for_training(
        self,
        candidate_individual_fragments_identifiers: list[int],
        accumulation_trial: int,
    ):
        """Assigns identities during test to individual fragments and rank them
        according to the score computed from the certainty of identification
        and the minimum distance traveled.

        Parameters
        ----------
        candidate_individual_fragments_identifiers : list
            List of fragment identifiers.
        """
        self.accumulation_strategy = "global"
        self.candidate_individual_fragments_identifiers = (
            candidate_individual_fragments_identifiers
        )
        self.reset_accumulation_variables()
        logging.debug("Accumulating by global strategy")
        for global_fragment in self.list_of_global_fragments.global_fragments:
            if not global_fragment.used_for_training:
                self.check_if_is_globally_acceptable_for_training(global_fragment)

        self.number_of_acceptable_global_fragments = sum(
            global_fragment.acceptable_for_training(self.accumulation_strategy)
            and not global_fragment.used_for_training
            for global_fragment in self.list_of_global_fragments.global_fragments
        )
        if accumulation_trial == 0:
            min_number_of_imgs_accumulated_to_start_partial_accumulation = (
                conf.MINIMUM_RATIO_OF_IMAGES_ACCUMULATED_GLOBALLY_TO_START_PARTIAL_ACCUMULATION
            )
        else:
            min_number_of_imgs_accumulated_to_start_partial_accumulation = 0
        if (
            self.number_of_acceptable_global_fragments == 0
            and self.ratio_accumulated_images
            > min_number_of_imgs_accumulated_to_start_partial_accumulation
            and self.ratio_accumulated_images < self.threshold_early_stop_accumulation
        ):
            logging.debug("Accumulating by partial strategy")
            self.accumulation_strategy = "partial"
            self.reset_accumulation_variables()
            for global_fragment in self.list_of_global_fragments.global_fragments:
                if not global_fragment.used_for_training:
                    self.check_if_is_partially_acceptable_for_training(global_fragment)
        elif (
            self.ratio_accumulated_images
            < min_number_of_imgs_accumulated_to_start_partial_accumulation
        ):
            logging.info(
                "The ratio of accumulated images is too small and a partial"
                " accumulation might fail."
            )

    def reset_non_acceptable_fragment(self, fragment: Fragment):
        """Resets the collection of non-acceptable fragments.

        Parameters
        ----------
        fragment : Fragment object
            Collection of images related to the same individual
        """
        if (
            fragment.identifier not in self.temporary_individual_fragments_used
            and fragment.identifier not in self.individual_fragments_used
        ):
            fragment.temporary_id = None
            fragment.acceptable_for_training = False

    def reset_non_acceptable_global_fragment(self, global_fragment: GlobalFragment):
        """Reset the flag for non-accpetable global fragments.

        Parameters
        ----------
        global_fragment : GlobalFragment object
            Collection of images relative to a part of the video in which all the animals are visible.
        """
        for fragment in global_fragment.individual_fragments:
            self.reset_non_acceptable_fragment(fragment)

    def check_if_is_globally_acceptable_for_training(
        self, global_fragment: GlobalFragment
    ):
        assert self.accumulation_strategy == "global"

        for fragment in global_fragment.individual_fragments:
            fragment.acceptable_for_training = True

        for fragment in global_fragment.individual_fragments:
            if fragment.identifier in self.candidate_individual_fragments_identifiers:
                if fragment.certainty < self.certainty_threshold:
                    # if the certainty of the individual fragment is not high enough
                    # we set the global fragment to be non-acceptable for training
                    self.reset_non_acceptable_global_fragment(global_fragment)
                    self.number_of_noncertain_global_fragments += 1
                    fragment.is_certain = False
                    break
                # if the certainty of the individual fragment is high enough
                fragment.is_certain = True
            elif fragment.identifier in self.individual_fragments_used:
                # if the individual fragment is not in the list of
                # candidates is because it has been assigned
                # and it is in the list of individual_fragments_used.
                # We set the certainty to 1. And we
                fragment.is_certain = True
            else:
                logging.warning(
                    "Individual fragment not in candidates or in used, this should"
                    " not happen"
                )
        # Compute identities if the global_fragment is certain
        if not global_fragment.acceptable_for_training("global"):
            return

        P1_array, index_individual_fragments_sorted_by_P1 = get_P1_array_and_argsort(
            global_fragment
        )
        # set to zero the P1 of the the identities of the individual
        # fragments that have been already used
        for index_individual_fragment, fragment in enumerate(
            global_fragment.individual_fragments
        ):
            if (
                fragment.identifier in self.individual_fragments_used
                or fragment.identifier in self.temporary_individual_fragments_used
            ):
                P1_array[index_individual_fragment, :] = 0.0
                P1_array[:, fragment.temporary_id] = 0.0
        # assign temporal identity to individual fragments by hierarchical P1
        for index_individual_fragment in index_individual_fragments_sorted_by_P1:
            fragment = global_fragment.individual_fragments[index_individual_fragment]
            assert isinstance(fragment, Fragment)
            if fragment.temporary_id is None:
                if p1_below_random(P1_array, index_individual_fragment, fragment):
                    fragment.P1_below_random = True
                    self.number_of_random_assigned_global_fragments += 1
                    self.reset_non_acceptable_global_fragment(global_fragment)
                    break

                temporary_id = np.argmax(P1_array[index_individual_fragment, :])
                if not fragment.check_consistency_with_coexistent_individual_fragments(
                    temporary_id
                ):
                    self.reset_non_acceptable_global_fragment(global_fragment)
                    fragment.non_consistent = True
                    self.number_of_nonconsistent_global_fragments += 1
                    break

                P1_array = set_fragment_temporary_id(
                    fragment, int(temporary_id), P1_array, index_individual_fragment
                )

        # Check if the global fragment is unique after assigning the identities
        if global_fragment.acceptable_for_training("global"):
            if not global_fragment.is_unique(self.n_animals):
                # set acceptable_for_training to False and temporary_id to
                # None for all the individual_fragments
                # that had not been accumulated before (i.e. not in
                # temporary_individual_fragments_used or individual_fragments_used)
                self.reset_non_acceptable_global_fragment(global_fragment)
                self.number_of_nonunique_global_fragments += 1
            else:
                global_fragment.accumulation_step = self.current_step
                self.temporary_individual_fragments_used.update(
                    fragment.identifier
                    for fragment in global_fragment.individual_fragments
                    if fragment.identifier not in self.individual_fragments_used
                )

    def check_if_is_partially_acceptable_for_training(
        self, global_fragment: GlobalFragment
    ):
        assert self.accumulation_strategy == "partial"
        for fragment in global_fragment.individual_fragments:
            fragment.acceptable_for_training = False

        for fragment in global_fragment.individual_fragments:
            # Check certainties of the individual fragme
            if fragment.identifier in self.candidate_individual_fragments_identifiers:
                if fragment.has_enough_accumulated_coexisting_fragments:
                    # Check if the more than half of the individual fragments
                    # that coexist with this one have being accumulated
                    if fragment.certainty < self.certainty_threshold:
                        # if the certainty of the individual fragment is not high enough
                        # we set the global fragment not to be acceptable for training
                        self.reset_non_acceptable_fragment(fragment)
                        self.number_of_noncertain_fragments += 1
                        fragment.is_certain = False
                    else:
                        # if the certainty of the individual fragment is high enough
                        fragment.is_certain = True
                        fragment.acceptable_for_training = True
                else:
                    self.reset_non_acceptable_fragment(fragment)
                    self.number_of_sparse_fragments += 1
            elif fragment.identifier in self.individual_fragments_used:
                # if the individual fragment is not in the list of candidates
                # is because it has been assigned
                # and it is in the list of individual_fragments_used.
                # We set the certainty to 1. And we
                fragment.is_certain = True
            else:
                logging.warning(
                    "Individual fragment not in candidates or in used, this should"
                    " not happen"
                )

        # Compute identities if the global_fragment is certain
        # get array of P1 values for the global fragment
        P1_array = np.asarray(
            [fragment.P1_vector for fragment in global_fragment.individual_fragments]
        )
        # get the maximum P1 of each individual fragment
        P1_max = P1_array.max(1)
        # logging.debug("P1 max: %s" %str(P1_max))
        # get the index position of the individual fragments ordered by
        # P1_max from max to min
        index_individual_fragments_sorted_by_P1 = np.argsort(P1_max)[::-1]
        # set to zero the P1 of the the identities of the individual
        # fragments that have been already used
        for index_individual_fragment, fragment in enumerate(
            global_fragment.individual_fragments
        ):
            if (
                fragment.identifier in self.individual_fragments_used
                or fragment.identifier in self.temporary_individual_fragments_used
            ):
                P1_array[index_individual_fragment, :] = 0.0
                P1_array[:, fragment.temporary_id] = 0.0

        # assign temporary identity to individual fragments by hierarchical P1
        for index_individual_fragment in index_individual_fragments_sorted_by_P1:
            fragment = global_fragment.individual_fragments[index_individual_fragment]
            assert isinstance(fragment, Fragment)  # for PyLance

            if fragment.temporary_id is None and fragment.acceptable_for_training:
                if (
                    P1_array[index_individual_fragment, :].max()
                    < 1.0 / fragment.number_of_images
                ):
                    fragment.P1_below_random = True
                    self.number_of_random_assigned_fragments += 1
                    self.reset_non_acceptable_fragment(fragment)
                else:
                    temporary_id = np.argmax(P1_array[index_individual_fragment, :])
                    if not fragment.check_consistency_with_coexistent_individual_fragments(
                        temporary_id
                    ):
                        self.reset_non_acceptable_fragment(fragment)
                        fragment.non_consistent = True
                        self.number_of_nonconsistent_fragments += 1
                    else:
                        fragment.acceptable_for_training = True
                        fragment.temporary_id = int(temporary_id)
                        P1_array[index_individual_fragment, :] = 0.0
                        P1_array[:, temporary_id] = 0.0

        # Check if the global fragment is unique after assigning the identities
        if not global_fragment.is_partially_unique:
            for fragment in global_fragment.individual_fragments:
                if fragment.temporary_id in global_fragment.duplicated_identities:
                    self.reset_non_acceptable_fragment(fragment)
                    self.number_of_nonunique_fragments += 1

        self.temporary_individual_fragments_used.update(
            fragment.identifier
            for fragment in global_fragment.individual_fragments
            if fragment.acceptable_for_training
            and fragment.identifier not in self.individual_fragments_used
        )
        self.number_of_acceptable_fragments += sum(
            bool(fragment.acceptable_for_training) and not fragment.used_for_training
            for fragment in global_fragment.individual_fragments
        )
        global_fragment.accumulation_step = self.current_step
        assert all(
            fragment.temporary_id is not None
            for fragment in global_fragment.individual_fragments
            if fragment.acceptable_for_training and fragment.is_an_individual
        )


def get_predictions_of_candidates_fragments(
    identification_model: Module,
    id_images_file_paths: list[Path],
    network_params: NetworkParams,
    list_of_fragments: ListOfFragments,
):
    """Get predictions of individual fragments that have been used to train the
    idCNN in an accumulation's iteration

    Parameters
    ----------
    net : ConvNetwork object
        network used to identify the animals
    video : Video object
        Object containing all the parameters of the video.
    fragments : list
        List of fragment objects

    Returns
    -------
    assigner._predictions  : nd.array
        predictions associated to each image organised by individual fragments
    assigner._softmax_probs : np.array
        softmax vector associated to each image organised by individual fragments
    np.cumsum(lengths)[:-1]  : nd.array
        cumulative sum of the number of images contained in every fragment
        (used to rebuild the collection of images per fragment after gathering
        predicions and softmax vectors from the gpu)
    candidate_individual_fragments_identifiers : list
        list of fragment identifiers
    """
    images = []
    lengths = []
    candidate_individual_fragments_identifiers: list[int] = []

    for fragment in list_of_fragments.individual_fragments:
        if not fragment.used_for_training:
            images += fragment.image_locations
            lengths.append(fragment.number_of_images)
            candidate_individual_fragments_identifiers.append(fragment.identifier)

    assert images
    images = load_id_images(id_images_file_paths, images)

    predictions, softmax_probs = get_predictions_identities(
        identification_model, images, network_params
    )

    assert sum(lengths) == len(predictions)
    return (
        predictions,
        softmax_probs,
        np.cumsum(lengths)[:-1],
        candidate_individual_fragments_identifiers,
    )


def get_P1_array_and_argsort(global_fragment: GlobalFragment):
    """Given a global fragment computes P1 for each of its individual
    fragments and returns a
    matrix of sorted indices according to P1

    Parameters
    ----------
    global_fragment : GlobalFragment object
        Collection of images relative to a part of the video in which all
        the animals are visible.

    Returns
    -------
    P1_array : nd.array
        P1 computed for every individual fragment in the global fragment
    index_individual_fragments_sorted_by_P1 : nd.array
        Argsort of P1 array of each individual fragment
    """
    # get array of P1 values for the global fragment
    P1_array = np.asarray(
        [fragment.P1_vector for fragment in global_fragment.individual_fragments]
    )
    # get the maximum P1 of each individual fragment
    P1_max = np.max(P1_array, axis=1)
    # get the index position of the individual fragments ordered by P1_max
    # from max to min
    index_individual_fragments_sorted_by_P1 = np.argsort(P1_max)[::-1]
    return P1_array, index_individual_fragments_sorted_by_P1


def p1_below_random(
    P1_array: np.ndarray, index_individual_fragment: np.ndarray, fragment: Fragment
):
    """Evaluate if a fragment has been assigned with a certainty lower than
    random (wrt the number of possible identities)

    Parameters
    ----------
    P1_array  : nd.array
        P1 vector of a fragment object
    index_individual_fragment  : nd.array
        Argsort of the P1 array of fragment
    fragment : Fragment
        Fragment object containing images associated with a single individual

    Returns
    -------
    p1_below_random_flag : bool
        True if a fragment has been identified with a certainty below random
    """
    return (
        P1_array[index_individual_fragment, :].max() < 1.0 / fragment.number_of_images
    )


def set_fragment_temporary_id(
    fragment: Fragment,
    temporary_id: int,
    P1_array: np.ndarray,
    index_individual_fragment: int,
):
    """Given a P1 array relative to a global fragment sets to 0 the row
    relative to fragment
    which is temporarily identified with identity temporary_id

    Parameters
    ----------
    fragment : Fragment
        Fragment object containing images associated with a single individual
    temporary_id : int
        temporary identifier associated to fragment
    P1_array  : nd.array
        P1 vector of fragment
    index_individual_fragment : int
        Index of fragment with respect to a global fragment in which it is
        contained

    Returns
    -------
    P1_array  : nd.array
        updated P1 array
    """
    fragment.temporary_id = int(temporary_id)
    P1_array[index_individual_fragment, :] = 0.0
    P1_array[:, temporary_id] = 0.0
    return P1_array



from collections.abc import Iterable

from qtpy.QtCore import Qt, Signal  # type: ignore
from qtpy.QtWidgets import (
    QHBoxLayout,
    QRadioButton,
    QScrollArea,
    QSpinBox,
    QVBoxLayout,
    QWidget,
)

from idtrackerai import Blob, Fragment


class MarkBlobs(QScrollArea):
    needToDraw = Signal()

    def __init__(self, parent: QWidget):
        main_layout = QVBoxLayout()
        super().__init__(parent)
        self.setWidgetResizable(True)
        wid = QWidget()
        wid.setLayout(main_layout)
        self.setWidget(wid)

        main_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self.individual = QRadioButton("Individuals")
        self.crossing = QRadioButton("Crossings")
        self.used_for_training_crossings = QRadioButton("Used for\ntraining crossings")
        self.used_for_training = QRadioButton("Used for\ntraining identification")
        self.accumulable = QRadioButton("Accumulable")
        self.not_accumulated = QRadioButton("Accumulable but\nnot accumulated")
        self.accumulated = QRadioButton("Accumulated at")
        self.accumulation_spinbox = QSpinBox()
        self.accumulation_spinbox.setSpecialValueText("any step")
        self.accumulation_spinbox.setRange(-1, 1000)
        self.accumulation_spinbox.setValue(-1)
        self.accumulation_spinbox.setPrefix("step ")
        self.forces_to_crossing = QRadioButton("Forced to be crossing")

        self.individual.toggled.connect(self.needToDraw.emit)
        self.used_for_training.toggled.connect(self.needToDraw.emit)
        self.used_for_training_crossings.toggled.connect(self.needToDraw.emit)
        self.accumulable.toggled.connect(self.needToDraw.emit)
        self.accumulated.toggled.connect(self.needToDraw.emit)
        self.accumulation_spinbox.valueChanged.connect(self.needToDraw.emit)
        self.accumulation_spinbox.valueChanged.connect(
            lambda: self.accumulated.setChecked(True)
        )
        self.not_accumulated.toggled.connect(self.needToDraw.emit)
        self.forces_to_crossing.toggled.connect(self.needToDraw.emit)

        main_layout.addWidget(self.individual)
        main_layout.addWidget(self.used_for_training)
        main_layout.addWidget(self.used_for_training_crossings)
        main_layout.addWidget(self.accumulable)
        accumulated_at = QHBoxLayout()
        accumulated_at.setAlignment(Qt.AlignmentFlag.AlignLeft)
        accumulated_at.addWidget(self.accumulated)
        accumulated_at.addWidget(self.accumulation_spinbox)
        main_layout.addLayout(accumulated_at)
        main_layout.addWidget(self.not_accumulated)
        main_layout.addWidget(self.forces_to_crossing)

    def __call__(
        self, blobs: list[Blob], fragments: list[Fragment] | None
    ) -> Iterable[Blob]:
        if not self.isVisible() or not self.isEnabled():
            return ()

        if self.individual.isChecked():
            return (blob for blob in blobs if blob.is_an_individual)

        if self.crossing.isChecked():
            return (blob for blob in blobs if blob.is_a_crossing)

        if self.used_for_training_crossings.isChecked():
            return (blob for blob in blobs if blob.used_for_training_crossings)

        if fragments is None:
            return ()

        if self.used_for_training.isChecked():
            return filter(
                lambda blob: fragments[blob.fragment_identifier].used_for_training,
                blobs,
            )

        if self.accumulable.isChecked():
            return filter(
                lambda blob: fragments[blob.fragment_identifier].accumulable, blobs
            )

        if self.accumulated.isChecked():
            step = self.accumulation_spinbox.value()
            if step == -1:  # Any step
                return filter(
                    lambda blob: fragments[blob.fragment_identifier].accumulation_step
                    is not None,
                    blobs,
                )

            return filter(
                lambda blob: fragments[blob.fragment_identifier].accumulation_step
                == step,
                blobs,
            )

        if self.not_accumulated.isChecked():
            return filter(
                lambda blob: fragments[blob.fragment_identifier].accumulable
                and fragments[blob.fragment_identifier].accumulation_step is None,
                blobs,
            )
        return ()



import numpy as np
from cv2 import fitEllipse
from qtpy.QtCore import Qt, Signal  # type: ignore
from qtpy.QtGui import QColor, QPainterPath
from qtpy.QtWidgets import (
    QCheckBox,
    QDialog,
    QGridLayout,
    QHBoxLayout,
    QListView,
    QListWidgetItem,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QToolButton,
    QVBoxLayout,
    QWidget,
)

from idtrackerai.utils import build_ROI_mask_from_list, get_vertices_from_label
from idtrackerai_GUI_tools import (
    CanvasMouseEvent,
    CanvasPainter,
    CustomList,
    build_ROI_patches_from_list,
)


class ROIWidget(QWidget):
    needToDraw = Signal()
    valueChanged = Signal(object)  # np.ndarray | None

    def __init__(self, parent):
        super().__init__()

        self.CheckBox = QCheckBox("Regions of interest")
        self.CheckBox.stateChanged.connect(self.CheckBox_changed)

        self.add = QToolButton()
        self.add.setText("Add")
        self.add.setCheckable(True)
        self.add.setVisible(False)

        self.list = CustomList()
        self.list.setVisible(False)

        self.list.setDefaultDropAction(Qt.DropAction.MoveAction)
        self.list.setMovement(QListView.Movement.Free)

        Controls_HBox = QHBoxLayout()
        Controls_HBox.addWidget(self.CheckBox)
        Controls_HBox.addWidget(self.add)

        layout = QVBoxLayout()
        layout.setSpacing(2)
        self.setLayout(layout)
        layout.addLayout(Controls_HBox)
        layout.addWidget(self.list)

        self.add.toggled.connect(self.add_clicked)
        self.list.ListChanged.connect(self.update_Patches)
        self.CheckBox.stateChanged.connect(self.update_Patches)
        self.list.ListChanged.connect(lambda: self.valueChanged.emit(self.getMask()))

        self.ROI_popup = ROI_PopUp(parent)
        self.list.newItemSelected.connect(self.paint_selected_polygon)
        self.mask_path = QPainterPath()
        self.clicked_points = []
        self.ListItem_clicked = False
        self.resolution_reduction = 1
        self.video_size = 1, 1

    def getValue(self) -> list[str] | None:
        return self.list.getValue() if self.CheckBox.isChecked() else None

    def getMask(self) -> np.ndarray | None:
        return build_ROI_mask_from_list(
            self.getValue(), self.resolution_reduction, *self.video_size
        )

    def CheckBox_changed(self, enabled):
        if self.add.isChecked():
            self.add.click()
        self.list.setVisible(enabled)
        self.add.setVisible(enabled)
        self.valueChanged.emit(self.getMask())

    def click_event(self, event: CanvasMouseEvent):
        if self.add.isChecked():
            if event.button == Qt.MouseButton.LeftButton:
                # Add clicked point
                self.clicked_points.append(event.xy_data)
            elif event.button == Qt.MouseButton.RightButton:
                # Remove nearest point
                if not self.clicked_points:
                    return
                distances = map(event.distance_to, self.clicked_points)
                index, dist = min(enumerate(distances), key=lambda x: x[1])
                if dist < 20 * event.zoom:  # 20 px threshold
                    self.clicked_points.pop(index)
            self.needToDraw.emit()

    def paint_selected_polygon(self, new: QListWidgetItem):
        if new:
            self.ListItem_clicked = True
            line = new.data(Qt.ItemDataRole.UserRole)
            self.clicked_points = list(
                map(tuple, get_vertices_from_label(line, close=True).astype(np.int32))
            )

        else:
            self.ListItem_clicked = False
            self.clicked_points.clear()
        self.needToDraw.emit()

    def add_clicked(self, checked):
        xy, self.clicked_points = self.clicked_points, []

        if checked:
            self.ROI_type = self.ROI_popup.exec(self.list.count() == 0)
            if self.ROI_type is None:
                self.add.setChecked(False)
            return

        if not xy:  # any drawn points
            return

        assert self.ROI_type is not None

        xy = np.asarray(xy) / self.resolution_reduction

        if self.ROI_type[2:9] == "Polygon":
            if len(xy) < 3:
                QMessageBox.warning(
                    self,
                    "ROI error",
                    "Polygons can only be defined with 3 points or more",
                )
            else:
                self.list.add_str(
                    f"{self.ROI_type} ["
                    + ", ".join([f"[{x:.1f}, {y:.1f}]" for x, y in xy])
                    + "]"
                )
        elif self.ROI_type[2:9] == "Ellipse":
            if len(xy) < 5:
                QMessageBox.warning(
                    self,
                    "ROI error",
                    "Ellipses can only be defined with 5 points"
                    "(exact fit) or more (approximated fit)",
                )
            else:
                center, axis, angle = fitEllipse(np.asarray(xy, dtype="f"))
                axis = axis[0] / 2.0, axis[1] / 2.0
                self.list.add_str(
                    f"{self.ROI_type} "
                    + "{"
                    + f"'center': [{center[0]:.0f}, {center[1]:.0f}], "
                    f"'axes': [{axis[0]:.0f}, {axis[1]:.0f}], 'angle': {angle:.0f}"
                    + "}"
                )

    def set_resolution_reduction(self, resolution_reduction: float):
        self.resolution_reduction = resolution_reduction
        self.update_Patches()
        self.valueChanged.emit(self.getMask())

    def set_video_size(self, video_size):
        self.video_size = video_size

    def update_Patches(self):
        self.mask_path = build_ROI_patches_from_list(
            self.getValue(), self.resolution_reduction, *self.video_size
        )

    def setValue(self, values: list[str] | str | None):
        if not values:
            return
        if isinstance(values, str):
            values = [values]
        self.CheckBox.setChecked(True)
        for value in values:
            self.list.add_str(value)

    def paint_on_canvas(self, painter: CanvasPainter):
        if not self.CheckBox.isChecked():
            return
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(QColor(255, 0, 0, 50))
        painter.drawPath(self.mask_path)

        painter.setBrush(Qt.BrushStyle.NoBrush)
        painter.setPenColor(QColor(0x32640A))
        if self.ListItem_clicked:
            painter.drawPolygonFromVertices(
                self.clicked_points, self.resolution_reduction
            )
        else:
            painter.setBrush(QColor(0x349650))
            for point in self.clicked_points:
                painter.drawBigPoint(*point)


class ROI_PopUp(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setWindowModality(Qt.WindowModality.ApplicationModal)
        self.setFixedSize(300, 100)
        self.setWindowTitle("Add ROI type")
        layout = QGridLayout()
        self.setLayout(layout)

        self.PP_button = QPushButton("+ Polygon")
        self.PE_button = QPushButton("+ Ellipse")
        self.NP_button = QPushButton("- Polygon")
        self.NE_button = QPushButton("- Ellipse")

        policy = QSizePolicy.Policy.Expanding
        self.PP_button.setSizePolicy(policy, policy)
        self.PE_button.setSizePolicy(policy, policy)
        self.NP_button.setSizePolicy(policy, policy)
        self.NE_button.setSizePolicy(policy, policy)

        self.PP_button.clicked.connect(self.clicked_event)
        self.PE_button.clicked.connect(self.clicked_event)
        self.NP_button.clicked.connect(self.clicked_event)
        self.NE_button.clicked.connect(self.clicked_event)

        layout.addWidget(self.PP_button, 0, 0)
        layout.addWidget(self.PE_button, 0, 1)
        layout.addWidget(self.NP_button, 1, 0)
        layout.addWidget(self.NE_button, 1, 1)

    def clicked_event(self):
        sender = self.sender()
        assert isinstance(sender, QPushButton)
        self.value = sender.text()
        self.accept()

    def exec(self, only_positive: bool = False) -> str | None:
        self.NP_button.setEnabled(not only_positive)
        self.NE_button.setEnabled(not only_positive)

        return self.value if super().exec() == QDialog.DialogCode.Accepted else None
